import json
from typing import Dict, Any, Optional

from kubernetes.client import V1Pod

from ....foundation.manager.stats.IMachineStats import IMachineStats


class KubernetesMachineStats(IMachineStats):
    """The class responsible to handle Kubernetes Machine statistics.

    Attributes:
        machine_api_object (V1Pod): The Kubernetes Pod associated with this statistics.
        lab_hash (str): The hash identifier of the network scenario of the Kubernetes Pod.
        name (str): The name of the device.
        pod_name (str): The Kubernetes Pod Name.
        image (str): The Image used for deploying the Kubernetes Pod.
        interfaces (str): The interfaces connected to this Kubernetes Pod.
        status (Optional[str]): The status of the Kubernetes Pod.
        assigned_node (Optional[str]): The cluster node assigned to this Kubernetes Pod.
    """
    __slots__ = ['machine_api_object', 'lab_hash', 'name', 'pod_name', 'image', 'status', 'assigned_node']

    def __init__(self, machine_api_object: V1Pod):
        self.machine_api_object: V1Pod = machine_api_object
        # Static Information
        self.lab_hash: str = machine_api_object.metadata.namespace
        self.name: str = machine_api_object.metadata.labels["name"]
        self.pod_name: str = machine_api_object.metadata.name

        container_statuses = machine_api_object.status.container_statuses
        self.image: str = container_statuses[0].image if container_statuses else "N/A"

        interfaces = []
        pod_networks_conf = json.loads(self.machine_api_object.metadata.annotations["k8s.v1.cni.cncf.io/networks"])
        for idx, network_conf in enumerate(pod_networks_conf):
            if 'kathara.link' in network_conf:
                interfaces.append(f"{idx}:{network_conf['kathara.link']}")
        self.interfaces: str = "-" if not interfaces else ", ".join(interfaces)

        # Dynamic Information
        self.status: Optional[str] = None
        self.assigned_node: Optional[str] = None

        self.update()

    def update(self) -> None:
        """Update dynamic statistics with the current ones.

        Returns:
            None
        """
        self.status = self._get_detailed_machine_status(self.machine_api_object)
        self.assigned_node = self.machine_api_object.spec.node_name

    @staticmethod
    def _get_detailed_machine_status(pod_api_object: V1Pod) -> str:
        """Return a string containing the Kubernetes Pod status.

        Args:
            pod_api_object (client.V1Pod): A Kubernetes Pod.

        Returns:
            str: A string containing the Kubernetes Pod status.
        """
        container_statuses = pod_api_object.status.container_statuses

        if not container_statuses:
            return pod_api_object.status.phase

        container_status = container_statuses[0].state

        string_status = None
        if container_status.terminated is not None:
            string_status = container_status.terminated.reason if container_status.terminated.reason is not None \
                else "Terminating"
        elif container_status.waiting is not None:
            string_status = container_status.waiting.reason

        # In case the status contains an error message, split it to the first ": " and take the left part
        return string_status.split(': ')[0] if string_status is not None else pod_api_object.status.phase

    def to_dict(self) -> Dict[str, Any]:
        """Transform statistics into a dict representation.

        Returns:
            Dict[str, Any]: Dict containing statistics.
        """
        return {
            "network_scenario_id": self.lab_hash,
            "name": self.name,
            "pod_name": self.pod_name,
            "image": self.image,
            "status": self.status,
            "assigned_node": self.assigned_node,
            "interfaces": self.interfaces,
        }

    def __repr__(self) -> str:
        return str(self.to_dict())

    def __str__(self) -> str:
        """Return a formatted string with the device statistics.

        Returns:
           str: a formatted string with the device statistics
        """
        formatted_stats = f"Network Scenario ID: {self.lab_hash}\n"
        formatted_stats += f"Device Name: {self.name}\n"
        formatted_stats += f"Pod Name: {self.pod_name}\n"
        formatted_stats += f"Image: {self.image}\n"
        formatted_stats += f"Status: {self.status}\n"
        formatted_stats += f"Assigned Node: {self.assigned_node}"
        formatted_stats += f"Interfaces: {self.interfaces}\n"

        return formatted_stats



import fcntl
import os
import signal
import struct
import sys
import termios
import tty
from typing import Any, Optional, Callable, Tuple

from ..core.IConsoleAdapter import IConsoleAdapter


def get_terminal_size_unix() -> Tuple[int, int]:
    """Get the current terminal size (Unix).

    Returns:
        Tuple[int, int]: a tuple containing (cols, rows).
    """
    buf = struct.pack("HHHH", 0, 0, 0, 0)
    res = fcntl.ioctl(sys.stdin.fileno(), termios.TIOCGWINSZ, buf)
    rows, cols, _, _ = struct.unpack("HHHH", res)
    return cols, rows


class UnixConsoleAdapter(IConsoleAdapter):
    """Unix-specific console adapter.

    Attributes:
        _stdin_fd (int): File descriptor for stdin.
        _stdout_fd (int): File descriptor for stdout.
        _orig_term_attrs (Optional[list]): Saved terminal attributes for restoring on exit_raw().
        _resize_handler_installed (bool): Whether resize watching is active.
        _stdin_reader_installed (bool): Whether stdin reader is installed.
    """

    __slots__ = ["_stdin_fd", "_stdout_fd", "_orig_term_attrs", "_resize_handler_installed", "_stdin_reader_installed"]

    def __init__(self) -> None:
        self._stdin_fd: int = sys.stdin.fileno()
        self._stdout_fd: int = sys.stdout.fileno()

        self._orig_term_attrs: Optional[list] = None

        self._resize_handler_installed: bool = False
        self._stdin_reader_installed: bool = False

    def enter_raw(self) -> None:
        """Put the terminal into raw mode.

        Returns:
            None
        """
        self._orig_term_attrs = termios.tcgetattr(self._stdin_fd)
        tty.setraw(self._stdin_fd)

    def exit_raw(self) -> None:
        """Restore the terminal to its previous mode (reverts "enter_raw").

        Returns:
            None
        """
        if self._orig_term_attrs is not None:
            termios.tcsetattr(self._stdin_fd, termios.TCSADRAIN, self._orig_term_attrs)
            self._orig_term_attrs = None

    def install_input_reader(self, loop: Any, on_bytes: Callable[[bytes], None], on_close: Callable[[], None]) -> None:
        """Start reading from stdin and route incoming bytes to callbacks (connected to the manager handler).

        Args:
            loop (Any): The asyncio event loop instance used to register readers/tasks.
            on_bytes (Callable[[bytes], None]): Callback to handle received input bytes.
            on_close (Callable[[], None]): Callback to handle stdin closing or errors.

        Returns:
            None
        """
        if self._stdin_reader_installed:
            return

        def _on_readable() -> None:
            try:
                data = os.read(self._stdin_fd, 4096)
            except OSError:
                on_close()
                return

            if not data:
                on_close()
                return

            on_bytes(data)

        loop.add_reader(self._stdin_fd, _on_readable)
        self._stdin_reader_installed = True

    def remove_input_reader(self, loop: Any) -> None:
        """Stop reading stdin previously configured by "install_input_reader".

        Args:
            loop (Any): The asyncio event loop instance used to deregister.

        Returns:
            None
        """
        if not self._stdin_reader_installed:
            return

        try:
            loop.remove_reader(self._stdin_fd)
        except Exception:
            pass

        self._stdin_reader_installed = False

    def write_stdout(self, data: bytes) -> None:
        """Write bytes to stdout.

        Args:
            data (bytes): Bytes to write to stdout.

        Returns:
            None
        """
        os.write(self._stdout_fd, data)

    def watch_resize(self, loop: Any, cb: Callable[[int, int], None]) -> None:
        """Start monitoring terminal resize events.

        Args:
            loop (Any): The asyncio event loop instance used to register signal handlers or callbacks.
            cb (Callable[[int, int], None]): Callback to handle the resize.

        Returns:
            None
        """

        def _emit_resize() -> None:
            cols, rows = get_terminal_size_unix()
            cb(cols, rows)

        try:
            loop.add_signal_handler(signal.SIGWINCH, _emit_resize)
        except Exception:
            signal.signal(signal.SIGWINCH, lambda *_: _emit_resize())

        self._resize_handler_installed = True
        _emit_resize()

    def unwatch_resize(self, loop: Any) -> None:
        """Stop monitoring terminal resize events.

        Args:
            loop (Any): The asyncio event loop instance used to deregister signal handlers or callbacks.

        Returns:
            None
        """
        if not self._resize_handler_installed:
            return

        try:
            loop.remove_signal_handler(signal.SIGWINCH)
        except Exception:
            pass

        self._resize_handler_installed = False



import argparse
import logging
from typing import List

from ... import utils
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...model.Lab import Lab
from ...parser.netkit.LabParser import LabParser
from ...strings import strings, wiki_description


class ConnectCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara connect',
            description=strings['connect'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show a help message and exit.'
        )

        group = self.parser.add_mutually_exclusive_group(required=False)

        group.add_argument(
            '-d', '--directory',
            help='Specify the folder containing the network scenario.',
        )
        group.add_argument(
            '-v', '--vmachine',
            dest="vmachine",
            action="store_true",
            help='The device has been started with vstart command.',
        )
        self.parser.add_argument(
            '--shell',
            required=False,
            help='Shell that should be used inside the device.'
        )
        self.parser.add_argument(
            '-l', '--logs',
            action="store_true",
            help='Print device startup logs before launching the shell.',
        )
        self.parser.add_argument(
            'machine_name',
            metavar='DEVICE_NAME',
            help='Name of the device to connect to.'
        )

    def run(self, current_path: str, argv: List[str]) -> int:
        self.parse_args(argv)
        args = self.get_args()

        if args['vmachine']:
            lab = Lab("kathara_vlab")
        else:
            lab_path = args['directory'].replace('"', '').replace("'", '') if args['directory'] else current_path
            lab_path = utils.get_absolute_path(lab_path)

            # Load custom 'kathara.conf' if it exists
            self._load_custom_configuration(lab_path)

            try:
                lab = LabParser.parse(lab_path)
            except (Exception, IOError):
                lab = Lab(None, path=lab_path)

        logging.debug(f"Executing `connect` command with hash `{lab.hash}`...")

        Kathara.get_instance().connect_tty(
            machine_name=args['machine_name'], lab_hash=lab.hash, shell=args['shell'], logs=args['logs']
        )

        return 0



import json
from typing import Any, Optional

from kubernetes.stream.ws_client import RESIZE_CHANNEL

from .....foundation.manager.terminal.core.ITerminalSession import ITerminalSession


class KubernetesWSTerminalSession(ITerminalSession):
    def __init__(self, handler: Any, client: Any = None) -> None:
        super().__init__(handler, client)

    def fileno(self) -> Optional[int]:
        """Return an OS-level file descriptor for the session, if available.

        Returns:
            Optional[int]: The file descriptor, or None if the session cannot be polled via fd-based readiness APIs.
        """
        return None

    def read(self, n: int = 4096) -> bytes:
        """Read up to n bytes from the session output stream.

        Args:
            n (int): Maximum number of bytes to read.

        Returns:
            bytes: Data read from the session.
        """
        if self._closed:
            return b""

        if not self._handler.is_open():
            raise Exception

        data = None
        try:
            if self._handler.peek_stdout():
                data = self._handler.read_stdout()
            elif self._handler.peek_stderr():
                data = self._handler.read_stderr()
        except Exception:
            return b""

        if data is None:
            return b""

        try:
            return data.encode("utf-8")
        except Exception:
            return b""

    def write(self, data: bytes) -> None:
        """Write bytes to the session input stream.

        Args:
            data (bytes): Data to send to the session.

        Returns:
            None
        """
        if self._closed:
            return

        try:
            self._handler.write_stdin(data)
        except Exception as e:
            raise e

    def resize(self, cols: int, rows: int) -> None:
        """Resize the session terminal dimensions.

        Args:
            cols (int): Terminal width in columns.
            rows (int): Terminal height in rows.

        Returns:
            None
        """
        if self._closed:
            return

        payload = json.dumps({"Height": rows, "Width": cols})
        try:
            self._handler.write_channel(RESIZE_CHANNEL, payload)
        except Exception as e:
            raise e

    def close(self) -> None:
        """Close the session and release resources.

        Returns:
            None
        """
        if self._closed:
            return

        self._closed = True
        self._handler.close()



import io
import json
import logging
from typing import Set, Dict, Generator, Any, List, Tuple, Optional, Union

from kubernetes import client
from kubernetes.client.rest import ApiException

from .KubernetesConfig import KubernetesConfig
from .KubernetesLink import KubernetesLink
from .KubernetesMachine import KubernetesMachine
from .KubernetesNamespace import KubernetesNamespace
from .KubernetesSecret import KubernetesSecret
from .exec_stream.KubernetesExecStream import KubernetesExecStream
from .stats.KubernetesLinkStats import KubernetesLinkStats
from .stats.KubernetesMachineStats import KubernetesMachineStats
from ... import utils
from ...exceptions import NotSupportedError, MachineNotFoundError, LinkNotFoundError, LabAlreadyExistsError, \
    InvocationError, LabNotFoundError
from ...foundation.manager.IManager import IManager
from ...model.Lab import Lab
from ...model.Link import Link
from ...model.Machine import Machine
from ...utils import pack_files_for_tar, check_required_single_not_none_var, check_single_not_none_var


class KubernetesManager(IManager):
    """Class responsible for interacting with Kubernetes API."""

    __slots__ = ['k8s_secret', 'k8s_namespace', 'k8s_machine', 'k8s_link']

    def __init__(self) -> None:
        KubernetesConfig.load_kube_config()

        self.k8s_namespace: KubernetesNamespace = KubernetesNamespace()
        self.k8s_secret: KubernetesSecret = KubernetesSecret()
        self.k8s_machine: KubernetesMachine = KubernetesMachine(self.k8s_namespace)
        self.k8s_link: KubernetesLink = KubernetesLink(self.k8s_namespace)

    def deploy_machine(self, machine: Machine) -> None:
        """Deploy a Kathara device.

        Args:
            machine (Kathara.model.Machine): A Kathara machine object.

        Returns:
            None

        Raises:
            LabNotFoundError: If the specified device is not associated to any network scenario.
            NonSequentialMachineInterfaceError: If there is a missing interface number in any device of the lab.
        """
        if not machine.lab:
            raise LabNotFoundError("Machine `%s` is not associated to a network scenario." % machine.name)

        machine.check()

        machine.lab.hash = machine.lab.hash.lower()

        self.k8s_namespace.create(machine.lab)
        self.k8s_secret.create(machine.lab)
        self.k8s_link.deploy_links(machine.lab, selected_links={x.link.name for x in machine.interfaces.values()})
        self.k8s_machine.deploy_machines(machine.lab, selected_machines={machine.name})

    def deploy_link(self, link: Link) -> None:
        """Deploy a Kathara collision domain.

        Args:
            link (Kathara.model.Link): A Kathara collision domain object.

        Returns:
            None

        Raises:
            LabNotFoundError: If the collision domain specified is not associated to any network scenario.
        """
        if not link.lab:
            raise LabNotFoundError(f"Collision domain `{link.name}` is not associated to a network scenario.")

        link.lab.hash = link.lab.hash.lower()

        self.k8s_namespace.create(link.lab)
        self.k8s_link.deploy_links(link.lab, selected_links={link.name})

    def deploy_lab(self, lab: Lab, selected_machines: Optional[Set[str]] = None,
                   excluded_machines: Optional[Set[str]] = None) -> None:
        """Deploy a Kathara network scenario.

        Args:
            lab (Kathara.model.Lab): A Kathara network scenario.
            selected_machines (Optional[Set[str]]): If not None, deploy only the specified devices.
            excluded_machines (Optional[Set[str]]): If not None, exclude devices from being deployed.

        Returns:
            None

        Raises:
            NonSequentialMachineInterfaceError: If there is a missing interface number in any device of the lab.
            MachineNotFoundError: If the specified devices are not in the network scenario specified.
            LabAlreadyExistsError: If a network scenario is deployed while it is terminating its execution.
            ApiError: If the Kubernetes APIs throw an exception.
            InvocationError: If both `selected_machines` and `excluded_machines` are specified.
        """
        lab.check_integrity()

        if selected_machines and excluded_machines:
            raise InvocationError(f"You can either select or exclude devices.")

        if selected_machines and not lab.has_machines(selected_machines):
            machines_not_in_lab = selected_machines - set(lab.machines.keys())
            raise MachineNotFoundError(f"The following devices are not in the network scenario: {machines_not_in_lab}.")

        if excluded_machines and not lab.has_machines(excluded_machines):
            machines_not_in_lab = excluded_machines - set(lab.machines.keys())
            raise MachineNotFoundError(f"The following devices are not in the network scenario: {machines_not_in_lab}.")

        # Kubernetes needs only lowercase letters for resources.
        # We force the hash to be lowercase
        lab.hash = lab.hash.lower()

        selected_links = None
        if selected_machines:
            selected_links = lab.get_links_from_machines(selected_machines)

        excluded_links = None
        if excluded_machines:
            # Get the links of remaining machines
            running_links = lab.get_links_from_machines(set(lab.machines.keys()) - excluded_machines)
            # Get the links of the excluded machines and get the diff with the running ones
            # The remaining are the ones to delete
            excluded_links = lab.get_links_from_machines(excluded_machines) - running_links

        self.k8s_namespace.create(lab)
        self.k8s_secret.create(lab)
        try:
            self.k8s_link.deploy_links(lab, selected_links=selected_links, excluded_links=excluded_links)

            self.k8s_machine.deploy_machines(
                lab, selected_machines=selected_machines, excluded_machines=excluded_machines
            )
        except ApiException as e:
            if e.status == 403 and 'Forbidden' in e.reason:
                raise LabAlreadyExistsError("Previous network scenario execution is still terminating. Please wait.")
            else:
                raise e

    def connect_machine_to_link(self, machine: Machine, link: Link, mac_address: Optional[str] = None) -> None:
        """Connect a Kathara device to a collision domain.

        Args:
            machine (Kathara.model.Machine): A Kathara machine object.
            link (Kathara.model.Link): A Kathara collision domain object.
            mac_address (Optional[str]): The MAC address to assign to the interface.

        Returns:
            None

        Raises:
            NotSupportedError: Unable to update a running device on Kubernetes.
        """
        raise NotSupportedError("Unable to update a running device.")

    def disconnect_machine_from_link(self, machine: Machine, link: Link) -> None:
        """Disconnect a Kathara device from a collision domain.

        Args:
            machine (Kathara.model.Machine): A Kathara machine object.
            link (Kathara.model.Link): The Kathara collision domain from which disconnect the device.

        Returns:
            None

        Raises:
            NotSupportedError: Unable to update a running device on Kubernetes.
        """
        raise NotSupportedError("Unable to update a running device.")

    def undeploy_machine(self, machine: Machine) -> None:
        """Undeploy a Kathara device.

        Args:
            machine (Kathara.model.Machine): A Kathara machine object.

        Returns:
            None

        Raises:
            LabNotFoundError: If the specified machine is not associated to a network scenario.
        """
        if not machine.lab:
            raise LabNotFoundError(f"Machine `{machine.name}` is not associated to a network scenario.")

        machine.lab.hash = machine.lab.hash.lower()

        # Get all current running machines (not Terminating), removing the one that we're undeploying
        running_machines = [running_machine for running_machine in
                            self.k8s_machine.get_machines_api_objects_by_filters(lab_hash=machine.lab.hash)
                            if 'Terminating' not in running_machine.status.phase and
                            running_machine.metadata.labels["name"] != machine.name]

        # From machines, save a set with all the attached networks (still needed)
        running_networks = set()
        for running_machine in running_machines:
            network_annotation = json.loads(running_machine.metadata.annotations["k8s.v1.cni.cncf.io/networks"])
            running_networks.update([net['name'] for net in network_annotation])

        # Difference between all networks of the machine to undeploy, and attached networks are the ones to delete
        machine_networks = {self.k8s_link.get_network_name(x.link.name) for x in machine.interfaces.values()}
        networks_to_delete = machine_networks - running_networks

        self.k8s_machine.undeploy(machine.lab.hash, selected_machines={machine.name})
        self.k8s_link.undeploy(machine.lab.hash, selected_links=networks_to_delete)

        if len(running_machines) <= 0:
            logging.debug("Waiting for namespace deletion...")

            self.k8s_namespace.undeploy(lab_hash=machine.lab.hash)

    def undeploy_link(self, link: Link) -> None:
        """Undeploy a Kathara collision domain.

        Args:
            link (Kathara.model.Link): A Kathara collision domain object.

        Returns:
            None

        Raises:
            LabNotFoundError: If the collision domain specified is not associated to any network scenario.
        """
        if not link.lab:
            raise LabNotFoundError(f"Collision domain `{link.name}` is not associated to a network scenario.")

        link.lab.hash = link.lab.hash.lower()

        network_name = self.k8s_link.get_network_name(link.name)

        # Get all current running machines (not Terminating)
        running_machines = [running_machine for running_machine in
                            self.k8s_machine.get_machines_api_objects_by_filters(lab_hash=link.lab.hash)
                            if 'Terminating' not in running_machine.status.phase]

        # From machines, save a set with all the attached networks (still needed)
        for running_machine in running_machines:
            network_annotation = json.loads(running_machine.metadata.annotations["k8s.v1.cni.cncf.io/networks"])
            # If the collision domain to undeploy is still used, do nothing
            if network_name in [net['name'] for net in network_annotation]:
                return

        self.k8s_link.undeploy(link.lab.hash, selected_links={network_name})

    def undeploy_lab(self, lab_hash: Optional[str] = None, lab_name: Optional[str] = None, lab: Optional[Lab] = None,
                     selected_machines: Optional[Set[str]] = None,
                     excluded_machines: Optional[Set[str]] = None) -> None:
        """Undeploy a Kathara network scenario.

        Args:
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab. If None, lab_name or lab should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab. If None, lab_hash or lab should be set.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name. If None, lab_hash or lab_name should be set.
            selected_machines (Optional[Set[str]]): If not None, undeploy only the specified devices.
            excluded_machines (Optional[Set[str]]): If not None, exclude devices from being undeployed.

        Returns:
            None

        Raises:
            InvocationError: If a running network scenario hash or name is not specified,
                or if both `selected_machines` and `excluded_machines` are specified.
        """
        check_required_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)
        if lab:
            lab_hash = lab.hash
        elif lab_name:
            lab_hash = utils.generate_urlsafe_hash(lab_name)

        if selected_machines and excluded_machines:
            raise InvocationError(f"You can either select or exclude devices.")

        lab_hash = lab_hash.lower()

        # When only some machines should be undeployed, special checks are required.
        if selected_machines or excluded_machines:
            # Get all current deployed networks and save only their name
            networks = self.k8s_link.get_links_api_objects_by_filters(lab_hash=lab_hash)
            all_networks = set([network["metadata"]["name"] for network in networks])

            # Get all current running machines (not Terminating)
            running_machines = [machine for machine in
                                self.k8s_machine.get_machines_api_objects_by_filters(lab_hash=lab_hash)
                                if 'Terminating' not in machine.status.phase
                                ]

            # From machines, save a set with all the attached networks (still needed)
            running_networks = set()
            for machine in running_machines:
                network_annotation = json.loads(machine.metadata.annotations["k8s.v1.cni.cncf.io/networks"])
                networks = [net['name'] for net in network_annotation]

                # To be running, you are not selected to undeploy, or you are excluded from undeploy
                if (selected_machines is not None and machine.metadata.labels["name"] not in selected_machines) or \
                        (excluded_machines is not None and machine.metadata.labels["name"] in excluded_machines):
                    running_networks.update(networks)

            # Difference between all networks and attached networks are the ones to delete
            networks_to_delete = all_networks - running_networks

            # Save only the fancy name of the machines
            running_machines = set([machine.metadata.labels["name"] for machine in running_machines])
        else:
            networks_to_delete = None
            running_machines = {}

        self.k8s_machine.undeploy(lab_hash, selected_machines=selected_machines, excluded_machines=excluded_machines)
        self.k8s_link.undeploy(lab_hash, selected_links=networks_to_delete)

        # Undeploy the namespace if:
        # 1- No machines are selected/excluded
        # 2- There are machines selected but the difference is empty
        # 3- There are machines excluded and the intersection is empty
        undeploy_namespace = (
                (not selected_machines and not excluded_machines) or
                (selected_machines and not (running_machines - selected_machines)) or
                (excluded_machines and not (running_machines & excluded_machines))
        )
        if undeploy_namespace:
            logging.debug("Waiting for namespace deletion...")

            self.k8s_namespace.undeploy(lab_hash=lab_hash)

    def wipe(self, all_users: bool = False) -> None:
        """Undeploy all the running network scenarios.

        Args:
            all_users (bool): If false, undeploy only the current user network scenarios. If true, undeploy the
                running network scenarios of all users.

        Returns:
            None
        """
        if all_users:
            logging.warning("User-specific options have no effect on Megalos.")

        self.k8s_namespace.wipe()

    def connect_tty(self, machine_name: str, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                    lab: Optional[Lab] = None, shell: str = None, logs: bool = False,
                    wait: Union[bool, Tuple[int, float]] = True) -> None:
        """Connect to a device in a running network scenario, using the specified shell.

        Args:
            machine_name (str): The name of the device to connect.
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab. If None, lab_name or lab should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab. If None, lab_hash or lab should be set.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name. If None, lab_hash or lab_name should be set.
            shell (str): The name of the shell to use for connecting.
            logs (bool): If True, print startup logs on stdout.
            wait (Union[bool, Tuple[int, float]]): If True, wait indefinitely until the end of the startup commands
                execution before connecting. If a tuple is provided, the first value indicates the number of retries
                before stopping waiting and the second value indicates the time interval to wait for each retry.
                Default is True.

        Returns:
            None

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
        """
        check_required_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)
        if lab:
            lab_hash = lab.hash
        elif lab_name:
            lab_hash = utils.generate_urlsafe_hash(lab_name)

        lab_hash = lab_hash.lower()

        self.k8s_machine.connect(lab_hash=lab_hash,
                                 machine_name=machine_name,
                                 shell=shell,
                                 logs=logs
                                 )

    def connect_tty_obj(self, machine: Machine, shell: str = None, logs: bool = False,
                        wait: Union[bool, Tuple[int, float]] = True) -> None:
        """Connect to a device in a running network scenario, using the specified shell.

        Args:
            machine (Machine): The device to connect.
            shell (str): The name of the shell to use for connecting.
            logs (bool): If True, print startup logs on stdout.
            wait (Union[bool, Tuple[int, float]]): If True, wait indefinitely until the end of the startup commands
                execution before connecting. If a tuple is provided, the first value indicates the number of retries
                before stopping waiting and the second value indicates the time interval to wait for each retry.
                Default is True.

        Returns:
            None

        Raises:
            LabNotFoundError: If the specified device is not associated to any network scenario.
        """
        if not machine.lab:
            raise LabNotFoundError(f"Device `{machine.name}` is not associated to a network scenario.")

        self.connect_tty(machine.name, lab=machine.lab, shell=shell, logs=logs, wait=wait)

    def exec(self, machine_name: str, command: Union[List[str], str], lab_hash: Optional[str] = None,
             lab_name: Optional[str] = None, lab: Optional[Lab] = None, wait: Union[bool, Tuple[int, float]] = False,
             stream: bool = True) -> Union[KubernetesExecStream, Tuple[bytes, bytes, int]]:
        """Exec a command on a device in a running network scenario.

        Args:
            machine_name (str): The name of the device to connect.
            command (Union[List[str], str]): The command to exec on the device.
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab. If None, lab_name or lab should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab. If None, lab_hash or lab should be set.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name. If None, lab_hash or lab_name should be set.
            wait (Union[bool, Tuple[int, float]]): If True, wait indefinitely until the end of the startup commands
                execution before executing the command. If a tuple is provided, the first value indicates the
                number of retries before stopping waiting and the second value indicates the time interval to wait
                for each retry. Default is False.
           stream (bool): If True, return a KubernetesExecStream object. If False,
                returns a tuple containing the complete stdout, the stderr, and the return code of the command.

        Returns:
            Union[KubernetesExecStream, Tuple[bytes, bytes, int]]: A KubernetesExecStream object or
            a tuple containing the stdout, the stderr and the return code of the command.

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
            MachineNotRunningError: If the specified device is not running.
            MachineBinaryError: If the binary of the command is not found.
        """
        check_required_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)
        if lab:
            lab_hash = lab.hash
        elif lab_name:
            lab_hash = utils.generate_urlsafe_hash(lab_name)

        lab_hash = lab_hash.lower()

        if wait:
            logging.warning("Wait option has no effect on Megalos.")

        return self.k8s_machine.exec(lab_hash, machine_name, command, stderr=True, tty=False, is_stream=stream)

    def exec_obj(self, machine: Machine, command: Union[List[str], str], wait: Union[bool, Tuple[int, float]] = False,
                 stream: bool = True) -> Union[KubernetesExecStream, Tuple[bytes, bytes, int]]:
        """Exec a command on a device in a running network scenario.

        Args:
            machine (Machine): The device to connect.
            command (Union[List[str], str]): The command to exec on the device.
            wait (Union[bool, Tuple[int, float]]): If True, wait indefinitely until the end of the startup commands
                execution before executing the command. If a tuple is provided, the first value indicates the
                number of retries before stopping waiting and the second value indicates the time interval to wait
                for each retry. Default is False.
            stream (bool): If True, return a KubernetesExecStream object. If False,
                returns a tuple containing the complete stdout, the stderr, and the return code of the command.

        Returns:
            Union[KubernetesExecStream, Tuple[bytes, bytes, int]]: A KubernetesExecStream object or
            a tuple containing the stdout, the stderr and the return code of the command.

        Raises:
            LabNotFoundError: If the specified device is not associated to any network scenario.
            MachineNotRunningError: If the specified device is not running.
            MachineBinaryError: If the binary of the command is not found.
        """
        if not machine.lab:
            raise LabNotFoundError(f"Device `{machine.name}` is not associated to a network scenario.")

        if wait:
            logging.warning("Wait option has no effect on Megalos.")

        return self.exec(machine.name, command, lab=machine.lab, wait=wait, stream=stream)

    def copy_files(self, machine: Machine, guest_to_host: Dict[str, Union[str, io.IOBase]]) -> None:
        """Copy files on a running device in the specified paths.

        Args:
            machine (Kathara.model.Machine): A running device object. It must have the api_object field populated.
            guest_to_host (Dict[str, Union[str, io.IOBase]]): A dict containing the device path as key and a
                fileobj to copy in path as value or a path to a file.

        Returns:
            None
        """
        tar_data = pack_files_for_tar(guest_to_host)

        self.k8s_machine.copy_files(machine.api_object, path="/", tar_data=tar_data)

    def retrieve_files(self, machine: Machine, src: str, dst: str) -> None:
        """Copy files from a running device path to the host.

        Args:
            machine (Kathara.model.Machine): A running device object. It must have the api_object field populated.
            src (str): The path of the file or folder to copy from the device.
            dst (str): The destination path on the host.

        Returns:
            None
        """
        self.k8s_machine.retrieve_files(machine.api_object, src, dst)

    def get_machine_api_object(self, machine_name: str, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                               lab: Optional[Lab] = None, all_users: bool = False) -> client.V1Pod:
        """Return the corresponding API object of a running device in a network scenario.

        Args:
            machine_name (str): The name of the device.
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab. If None, lab_name or lab should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab. If None, lab_hash or lab should be set.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name. If None, lab_hash or lab_name should be set.
            all_users (bool): If True, return information about devices of all users.

        Returns:
            client.V1Pod: A Kubernetes Pod.

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
            MachineNotFoundError: If the device is not found.
        """
        check_required_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)
        if lab:
            lab_hash = lab.hash
        elif lab_name:
            lab_hash = utils.generate_urlsafe_hash(lab_name)

        lab_hash = lab_hash.lower()

        if all_users:
            logging.warning("User-specific options have no effect on Megalos.")

        pods = self.k8s_machine.get_machines_api_objects_by_filters(lab_hash=lab_hash, machine_name=machine_name)
        if pods:
            return pods.pop()

        raise MachineNotFoundError(f"Device {machine_name} not found.")

    def get_machines_api_objects(self, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                                 lab: Optional[Lab] = None, all_users: bool = False) -> List[client.V1Pod]:
        """Return API objects of running devices.

        Args:
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name.
            all_users (bool): If True, return information about devices of all users.

        Returns:
            List[client.V1Pod]: Kubernetes Pod objects of devices.
        """
        check_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)
        if lab:
            lab_hash = lab.hash
        elif lab_name:
            lab_hash = utils.generate_urlsafe_hash(lab_name)

        lab_hash = lab_hash.lower() if lab_hash else None

        if all_users:
            logging.warning("User-specific options have no effect on Megalos.")

        return self.k8s_machine.get_machines_api_objects_by_filters(lab_hash=lab_hash)

    def get_link_api_object(self, link_name: str, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                            lab: Optional[Lab] = None, all_users: bool = False) -> Any:
        """Return the corresponding API object of a collision domain in a network scenario.

        Args:
            link_name (str): The name of the collision domain.
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab. If None, lab_name or lab should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab. If None, lab_hash or lab should be set.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name. If None, lab_hash or lab_name should be set.
            all_users (bool): If True, return information about collision domains of all users.

        Returns:
            Any: Kubernetes API object of the network.

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
            LinkNotFoundError: If the collision domain is not found.
        """
        check_required_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)
        if lab:
            lab_hash = lab.hash
        elif lab_name:
            lab_hash = utils.generate_urlsafe_hash(lab_name)

        lab_hash = lab_hash.lower()

        if all_users:
            logging.warning("User-specific options have no effect on Megalos.")

        networks = self.k8s_link.get_links_api_objects_by_filters(lab_hash=lab_hash, link_name=link_name)
        if networks:
            return networks.pop()

        raise LinkNotFoundError(f"Collision Domain {link_name} not found.")

    def get_links_api_objects(self, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                              lab: Optional[Lab] = None, all_users: bool = False) -> List[Any]:
        """Return API objects of collision domains in a network scenario.

        Args:
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name.
            all_users (bool): If True, return information about collision domains of all users.

        Returns:
            List[Any]: Kubernetes API objects of networks.
        """
        check_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)
        if lab:
            lab_hash = lab.hash
        elif lab_name:
            lab_hash = utils.generate_urlsafe_hash(lab_name)

        lab_hash = lab_hash.lower() if lab_hash else None

        if all_users:
            logging.warning("User-specific options have no effect on Megalos.")

        return self.k8s_link.get_links_api_objects_by_filters(lab_hash=lab_hash)

    def get_lab_from_api(self, lab_hash: str = None, lab_name: str = None) -> Lab:
        """Return the network scenario (specified by the hash or name), building it from API objects.

        Args:
            lab_hash (str): The hash of the network scenario. Can be used as an alternative to lab_name.
            lab_name (str): The name of the network scenario. Can be used as an alternative to lab_name.

        Returns:
            Lab: The built network scenario.

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
        """
        if not lab_hash and not lab_name:
            raise InvocationError("You must specify a running network scenario hash or name.")

        if lab_name:
            reconstructed_lab = Lab(lab_name)
        else:
            reconstructed_lab = Lab("reconstructed_lab")
            reconstructed_lab.hash = lab_hash

        lab_pods = self.get_machines_api_objects(lab_hash=reconstructed_lab.hash)
        lab_networks = dict(
            map(lambda x: (x['metadata']['name'], x), self.get_links_api_objects(lab_hash=reconstructed_lab.hash))
        )

        for pod in lab_pods:
            device = reconstructed_lab.get_or_new_machine(pod.metadata.labels["name"])
            device.api_object = pod

            # Rebuild device metas
            # NOTE: "privileged" and "bridged" are not supported on Megalos
            # NOTE: We cannot rebuild "sysctls", "exec", "ipv6" and "num_terms" meta.
            container = pod.spec.containers[0]
            device.add_meta("image", container.image)
            device.add_meta("shell", self.k8s_machine.get_env_var_value_from_pod(pod, "_MEGALOS_SHELL"))

            if container.resources.limits and 'memory' in container.resources.limits:
                device.add_meta("mem", container.resources.limits['memory'].upper())

            # Reconvert mcpus to a value passed by the user
            if container.resources.limits and 'cpu' in container.resources.limits:
                device.add_meta("cpu", int(container.resources.limits['cpu'].replace('m', '')) / 1000)

            for env in container.env:
                if env.name != "_MEGALOS_SHELL":
                    device.meta["envs"][env.name] = env.value

            # Reconvert ports to the device format
            if container.ports:
                for port in container.ports:
                    device.meta["ports"][(port.host_port, port.protocol.lower())] = port.container_port

            for network_conf in json.loads(pod.metadata.annotations["k8s.v1.cni.cncf.io/networks"]):
                network = lab_networks[network_conf['name']]
                link = reconstructed_lab.get_or_new_link(network['metadata']['labels']['name'])
                link.api_object = network

                iface_mac_addr = None
                if "mac" in network_conf:
                    iface_mac_addr = network_conf['mac']

                device.add_interface(link, mac_address=iface_mac_addr)

        return reconstructed_lab

    def update_lab_from_api(self, lab: Lab) -> None:
        """Update the passed network scenario from API objects.

        Args:
            lab (Lab): The network scenario to update.

        Raises:
            NotSupportedError: Unable to update a running network scenario on Kubernetes.
        """
        raise NotSupportedError("Unable to update a running network scenario.")

    def get_machines_stats(self, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                           lab: Optional[Lab] = None, machine_name: str = None, all_users: bool = False) \
            -> Generator[Dict[str, KubernetesMachineStats], None, None]:
        """Return information about the running devices.

        Args:
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name.
            machine_name (str): If specified return all the devices with machine_name.
            all_users (bool): If True, return information about the device of all users.

        Returns:
              Generator[Dict[str, KubernetesMachineStats], None, None]: A generator containing dicts that has API Object
                identifier as keys and KubernetesMachineStats objects as values.
        """
        check_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)
        if lab:
            lab_hash = lab.hash
        elif lab_name:
            lab_hash = utils.generate_urlsafe_hash(lab_name)

        lab_hash = lab_hash.lower() if lab_hash else None

        if all_users:
            logging.warning("User-specific options have no effect on Megalos.")

        return self.k8s_machine.get_machines_stats(lab_hash=lab_hash, machine_name=machine_name)

    def get_machine_stats(self, machine_name: str, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                          lab: Optional[Lab] = None, all_users: bool = False) \
            -> Generator[Optional[KubernetesMachineStats], None, None]:
        """Return information of the specified device in a specified network scenario.

        Args:
            machine_name (str): The name of the device for which statistics are requested.
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab. If None, lab_name or lab should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab. If None, lab_hash or lab should be set.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name. If None, lab_hash or lab_name should be set.
            all_users (bool): If True, search the device among all the users devices.

        Returns:
            Generator[Optional[KubernetesMachineStats], None, None]: A generator containing the KubernetesMachineStats
            object with the device info. Returns None if the device is not found.

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
        """
        check_required_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)
        if lab:
            lab_hash = lab.hash
        elif lab_name:
            lab_hash = utils.generate_urlsafe_hash(lab_name)

        lab_hash = lab_hash.lower()

        machines_stats = self.get_machines_stats(lab_hash=lab_hash, machine_name=machine_name)
        machines_stats_next = next(machines_stats)
        if machines_stats_next:
            (_, machine_stats) = machines_stats_next.popitem()
            yield machine_stats
        else:
            yield None

    def get_machine_stats_obj(self, machine: Machine, all_users: bool = False) \
            -> Generator[Optional[KubernetesMachineStats], None, None]:
        """Return information of the specified device in a specified network scenario.

        Args:
            machine (Machine): The device for which statistics are requested.
            all_users (bool): If True, search the device among all the users devices.

        Returns:
            Generator[Optional[IMachineStats], None, None]: A generator containing the IMachineStats object
            with the device info. Returns None if the device is not found.

        Raises:
            LabNotFoundError: If the specified device is not associated to any network scenario.
            MachineNotRunningError: If the specified device is not running.
        """
        if not machine.lab:
            raise LabNotFoundError("Device `%s` is not associated to a network scenario." % machine.name)

        return self.get_machine_stats(machine.name, lab=machine.lab, all_users=all_users)

    def get_links_stats(self, lab_hash: Optional[str] = None, lab_name: Optional[str] = None, lab: Optional[Lab] = None,
                        link_name: str = None, all_users: bool = False) \
            -> Generator[Dict[str, KubernetesLinkStats], None, None]:
        """Return information about deployed networks.

        Args:
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name.
           link_name (str): If specified return all the networks with link_name.
           all_users (bool): If True, return information about the networks of all users.

        Returns:
             Generator[Dict[str, KubernetesLinkStats], None, None]: A generator containing dicts that has API Object
                identifier as keys and KubernetesLinkStats objects as values.
        """
        check_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)
        if lab:
            lab_hash = lab.hash
        elif lab_name:
            lab_hash = utils.generate_urlsafe_hash(lab_name)

        lab_hash = lab_hash.lower() if lab_hash else None

        if all_users:
            logging.warning("User-specific options have no effect on Megalos.")

        return self.k8s_link.get_links_stats(lab_hash=lab_hash, link_name=link_name)

    def get_link_stats(self, link_name: str, lab_hash: Optional[str] = None, lab_name: Optional[str] = None,
                       lab: Optional[Lab] = None, all_users: bool = False) \
            -> Generator[Optional[KubernetesLinkStats], None, None]:
        """Return information of the specified deployed network in a specified network scenario.

        Args:
            link_name (str): The name of the collision domain for which statistics are requested.
            lab_hash (Optional[str]): The hash of the network scenario.
                Can be used as an alternative to lab_name and lab. If None, lab_name or lab should be set.
            lab_name (Optional[str]): The name of the network scenario.
                Can be used as an alternative to lab_hash and lab. If None, lab_hash or lab should be set.
            lab (Optional[Kathara.model.Lab]): The network scenario object.
                Can be used as an alternative to lab_hash and lab_name. If None, lab_hash or lab_name should be set.
            all_users (bool): If True, return information about the networks of all users.

        Returns:
            Generator[Optional[KubernetesLinkStats], None, None]: A generator containing the KubernetesLinkStats object
            with the network info. Returns None if the network is not found.

        Raises:
            InvocationError: If a running network scenario hash or name is not specified.
        """
        check_required_single_not_none_var(lab_hash=lab_hash, lab_name=lab_name, lab=lab)
        if lab:
            lab_hash = lab.hash
        elif lab_name:
            lab_hash = utils.generate_urlsafe_hash(lab_name)

        lab_hash = lab_hash.lower()

        links_stats = self.get_links_stats(lab_hash=lab_hash, link_name=link_name, all_users=all_users)
        links_stats_next = next(links_stats)
        if links_stats_next:
            (_, link_stats) = links_stats_next.popitem()
            yield link_stats
        else:
            yield None

    def get_link_stats_obj(self, link: Link, all_users: bool = False) \
            -> Generator[Optional[KubernetesLinkStats], None, None]:
        """Return information of the specified deployed network in a specified network scenario.

        Args:
            link (Link): The collision domain for which statistics are requested.
            all_users (bool): If True, return information about the networks of all users.

        Returns:
            Generator[Optional[ILinkStats], None, None]: A generator containing the ILinkStats object
            with the network info. Returns None if the network is not found.

        Raises:
            LabNotFoundError: If the specified device is not associated to any network scenario.
        """
        if not link.lab:
            raise LabNotFoundError(f"Link `{link.name}` is not associated to a network scenario.")

        return self.get_link_stats(link.name, lab=link.lab, all_users=all_users)

    def check_image(self, image_name: str) -> None:
        """Useless. The Check of the image is delegated to Kubernetes.

        Args:
            image_name (str): The name of the image

        Returns:
            None
        """
        # Delegate the image check to Kubernetes
        return

    def get_release_version(self) -> str:
        """Return the current manager version.

        Returns:
            str: The current manager version.
        """
        return client.VersionApi().get_code().git_version

    @staticmethod
    def get_formatted_manager_name() -> str:
        """Return a formatted string containing the current manager name.

        Returns:
            str: A formatted string containing the current manager name.
        """
        return "Kubernetes (Megalos)"



import logging
import os
import re
from multiprocessing.dummy import Pool
from typing import List, Union, Dict, Generator, Set, Optional

import docker
import docker.models.networks
from docker import DockerClient
from docker import types

from .DockerPlugin import DockerPlugin
from .stats.DockerLinkStats import DockerLinkStats
from ... import utils
from ...event.EventDispatcher import EventDispatcher
from ...exceptions import PrivilegeError, InvocationError
from ...model.ExternalLink import ExternalLink
from ...model.Lab import Lab
from ...model.Link import BRIDGE_LINK_NAME, Link
from ...os.Networking import Networking
from ...setting.Setting import Setting
from ...types import SharedCollisionDomainsOption


class DockerLink(object):
    """The class responsible for deploying Kathara collision domains as Docker networks and interact with them."""
    __slots__ = ['client', 'docker_plugin']

    def __init__(self, client: DockerClient, docker_plugin: DockerPlugin) -> None:
        self.client: DockerClient = client
        self.docker_plugin: DockerPlugin = docker_plugin

    def deploy_links(self, lab: Lab, selected_links: Set[str] = None, excluded_links: Set[str] = None) -> None:
        """Deploy all the network scenario collision domains as Docker networks.

        Args:
            lab (Kathara.model.Lab.Lab): A Kathara network scenario.
            selected_links (Set[str]): A set containing the name of the collision domains to deploy.
            excluded_links (Set[str]): A set containing the name of the collision domains to exclude.

        Returns:
            None

        Raises:
            InvocationError: If both `selected_links` and `excluded_links` are specified.
        """
        if selected_links and excluded_links:
            raise InvocationError(f"You can either specify `selected_links` or `excluded_links`.")

        links = lab.links.items()
        if selected_links:
            links = {
                k: v for k, v in links if k in selected_links
            }.items()
        elif excluded_links:
            links = {
                k: v for k, v in links if k not in excluded_links
            }.items()

        if len(links) > 0:
            pool_size = utils.get_pool_size()
            items = utils.chunk_list(links, pool_size)

            EventDispatcher.get_instance().dispatch("links_deploy_started", items=links)

            with Pool(pool_size) as links_pool:
                for chunk in items:
                    links_pool.map(func=self._deploy_link, iterable=chunk)

            EventDispatcher.get_instance().dispatch("links_deploy_ended")

        # Create a docker bridge link in the lab object and assign the Docker Network object associated to it.
        docker_bridge = self.get_docker_bridge()
        link = lab.get_or_new_link(BRIDGE_LINK_NAME)
        link.api_object = docker_bridge

    def _deploy_link(self, link_item: (str, Link)) -> None:
        """Deploy the collision domain contained in the link_item as a Docker network.

        Args:
            link_item (Tuple[str, Link]): A tuple composed by the name of the collision domain and a Link object

        Returns:
            None
        """
        (_, link) = link_item

        if link.name == BRIDGE_LINK_NAME:
            return

        self.create(link)

        EventDispatcher.get_instance().dispatch("link_deployed", item=link)

    def create(self, link: Link) -> None:
        """Create a Docker network representing the collision domain object and assign it to link.api_object.

        It also connects external collision domains, if present.

        Args:
            link (Kathara.model.Link.Link): A Kathara collision domain.

        Returns:
            None

        Raises:
            OSError: If the link is attached to external interfaces and the host OS is not LINUX.
            PrivilegeError: If the link is attached to external interfaces and the user does not have root privileges.
        """
        # Reserved name for bridged connections, ignore.
        if link.name == BRIDGE_LINK_NAME:
            return

        # If a network with the same name exists, return it instead of creating a new one.
        filter_lab_hash = None
        filter_user = None
        if Setting.get_instance().shared_cds == SharedCollisionDomainsOption.NOT_SHARED:
            filter_lab_hash = link.lab.hash
        if Setting.get_instance().shared_cds != SharedCollisionDomainsOption.USERS:
            filter_user = utils.get_current_user_name()

        networks = self.get_links_api_objects_by_filters(
            link_name=link.name, lab_hash=filter_lab_hash, user=filter_user
        )
        if networks:
            link.api_object = networks.pop()
        else:
            network_ipam_config = docker.types.IPAMConfig(driver='null')

            link_name = self.get_network_name(link)
            additional_labels = {}
            if Setting.get_instance().shared_cds != SharedCollisionDomainsOption.USERS:
                additional_labels["user"] = utils.get_current_user_name()
            if Setting.get_instance().shared_cds == SharedCollisionDomainsOption.NOT_SHARED:
                additional_labels["lab_hash"] = link.lab.hash

            link.api_object = self.client.networks.create(
                name=link_name,
                driver=f"{Setting.get_instance().network_plugin}:{utils.get_architecture()}",
                check_duplicate=True,
                ipam=network_ipam_config,
                labels={
                    "name": link.name,
                    "app": "kathara",
                    "external": ";".join([x.get_full_name() for x in link.external]),
                    **additional_labels
                }
            )

        if link.external:
            logging.debug("External Interfaces required, connecting them...")
            self._attach_external_interfaces(link.external, link.api_object)

    def undeploy(self, lab_hash: str, selected_links: Optional[Set[str]] = None) -> None:
        """Undeploy all the collision domains of the scenario specified by lab_hash.

        Args:
            lab_hash (str): The hash of the network scenario to undeploy.
            selected_links (Set[str]): If specified, delete only the collision domains contained in the set.

        Returns:
            None
        """
        networks = self.get_links_api_objects_by_filters(lab_hash=lab_hash)
        if selected_links:
            networks = [item for item in networks if item.attrs["Labels"]["name"] in selected_links]

        for item in networks:
            item.reload()
        networks = [item for item in networks if len(item.containers) <= 0]

        if len(networks) > 0:
            pool_size = utils.get_pool_size()
            items = utils.chunk_list(networks, pool_size)

            EventDispatcher.get_instance().dispatch("links_undeploy_started", items=networks)

            with Pool(pool_size) as links_pool:
                for chunk in items:
                    links_pool.map(func=self._undeploy_link, iterable=chunk)

            EventDispatcher.get_instance().dispatch("links_undeploy_ended")

    def wipe(self, user: str = None) -> None:
        """Undeploy all the Docker networks of the specified user. If user is None, it undeploy all the Docker networks.

        Args:
            user (str): The name of a current user on the host

        Returns:
            None
        """
        user_label = user if Setting.get_instance().shared_cds != SharedCollisionDomainsOption.USERS else None
        networks = self.get_links_api_objects_by_filters(user=user_label)
        for item in networks:
            item.reload()
        networks = [item for item in networks if len(item.containers) <= 0]

        pool_size = utils.get_pool_size()
        items = utils.chunk_list(networks, pool_size)

        with Pool(pool_size) as links_pool:
            for chunk in items:
                links_pool.map(func=self._undeploy_link, iterable=chunk)

    def _undeploy_link(self, network: docker.models.networks.Network) -> None:
        """Undeploy a Docker network.

        Args:
            network (docker.models.networks.Network): The Docker network to undeploy.

        Returns:
            None
        """
        self._delete_link(network)

        EventDispatcher.get_instance().dispatch("link_undeployed", item=network)

    def get_docker_bridge(self) -> Union[None, docker.models.networks.Network]:
        """Return the Docker bridged network.

        Returns:
            Union[None, docker.models.networks.Network]: The Docker bridged network if exists, else None
        """
        bridge_list = self.client.networks.list(names="bridge")
        return bridge_list.pop() if bridge_list else None

    def get_links_api_objects_by_filters(self, lab_hash: str = None, link_name: str = None, user: str = None) -> \
            List[docker.models.networks.Network]:
        """Return the Docker networks specified by lab_hash and user.

        Args:
            lab_hash (str): The hash of a network scenario. If specified, return all the networks in the scenario.
            link_name (str): The name of a network. If specified, return the specified network of the scenario.
            user (str): The name of a user on the host. If specified, return only the networks of the user.

        Returns:
            List[docker.models.networks.Network]: A list of Docker networks.
        """
        filters = {"label": ["app=kathara"]}
        if user:
            filters["label"].append(f"user={user}")
        if lab_hash:
            filters["label"].append(f"lab_hash={lab_hash}")
        if link_name:
            filters["label"].append(f"name={link_name}")

        return self.client.networks.list(filters=filters, greedy=True)

    def get_links_stats(self, lab_hash: str = None, link_name: str = None, user: str = None) -> \
            Generator[Dict[str, DockerLinkStats], None, None]:
        """Return a generator containing the Docker networks' stats.

        Args:
           lab_hash (str): The hash of a network scenario. If specified, return all the stats of the networks in the
           scenario.
           link_name (str): The name of a device. If specified, return the specified network stats.
           user (str): The name of a user on the host. If specified, return only the stats of the specified user.

        Returns:
           Generator[Dict[str, DockerMachineStats], None, None]: A generator containing network names as keys and
           DockerLinkStats as values.

        Raises:
            PrivilegeError: If user param is None and the user does not have root privileges.
        """
        if user is None and not utils.is_admin():
            raise PrivilegeError("You must be root to get networks statistics of all users.")

        networks_stats = {}

        def load_link_stats(network):
            if network.name not in networks_stats:
                networks_stats[network.name] = DockerLinkStats(network)

        while True:
            networks = self.get_links_api_objects_by_filters(lab_hash=lab_hash, link_name=link_name, user=user)
            if not networks:
                yield dict()

            pool_size = utils.get_pool_size()
            items = utils.chunk_list(networks, pool_size)
            with Pool(pool_size) as links_pool:
                for chunk in items:
                    links_pool.map(func=load_link_stats, iterable=chunk)

            networks_to_remove = []
            for network_id, network_stats in networks_stats.items():
                try:
                    network_stats.update()
                except StopIteration:
                    networks_to_remove.append(network_id)
                    continue

            for k in networks_to_remove:
                networks_stats.pop(k, None)

            yield networks_stats

    def _delete_link(self, network: docker.models.networks.Network) -> None:
        """Delete a Docker network.

        Args:
            network (docker.models.networks.Network): A Docker network.

        Raises:
            PrivilegeError: If you are not root while deleting an external VLAN Interface.
        """
        external_label = network.attrs['Labels']["external"]
        external_links = external_label.split(";") if external_label else None

        if external_links:
            self._delete_external_interfaces(external_links, network)

        network.remove()

    def _attach_external_interfaces(self, external_links: List[ExternalLink],
                                    network: docker.models.networks.Network) -> None:
        """Attach external collision domains to a Docker network.

        Args:
            external_links (List[Kathara.model.ExternalLink]): A list of Kathara external collision domains.
                They are used to create collision domains attached to a host interface.
            network (docker.models.networks.Network): A Docker network.

        Returns:
            None

        Raises:
            OSError: If the link is attached to external interfaces and the host OS is not LINUX.
            PrivilegeError: If the link is attached to external interfaces and the user does not have root privileges.
        """
        if not (utils.is_platform(utils.LINUX) or utils.is_platform(utils.LINUX2)):
            raise OSError("External collision domains available only on Linux systems.")

        if not utils.is_admin():
            raise PrivilegeError("You must be root in order to use external collision domains.")

        for external_link in external_links:
            (name, vlan) = external_link.get_name_and_vlan()
            interface_index = Networking.get_or_new_interface(external_link.interface, name, vlan)
            bridge_name = self._get_bridge_name(network)

            def vde_attach():
                plugin_pid = self.docker_plugin.plugin_pid()
                switch_path = os.path.join(self.docker_plugin.plugin_store_path(), bridge_name)
                Networking.attach_interface_ns(external_link.get_full_name(), interface_index, switch_path, plugin_pid)

            def bridge_attach():
                Networking.attach_interface_bridge(interface_index, bridge_name)

            self.docker_plugin.exec_by_version(vde_attach, bridge_attach)

    def _delete_external_interfaces(self, external_links: List[str], network: docker.models.networks.Network) -> None:
        """Remove external collision domains from a Docker network.

        Args:
            external_links (List[Kathara.model.ExternalLink]): A list of Kathara external collision domains.
            network (docker.models.networks.Network): A Docker network.

        Returns:
            None

        Raises:
            OSError: If the link is attached to external interfaces and the host OS is not LINUX.
            PrivilegeError: If the link is attached to external interfaces and the user does not have root privileges.
        """
        if not (utils.is_platform(utils.LINUX) or utils.is_platform(utils.LINUX2)):
            raise OSError("External collision domains available only on Linux systems.")

        if not utils.is_admin():
            raise PrivilegeError("You must be root in order to use external collision domains.")

        for external_link in external_links:
            def vde_delete():
                plugin_pid = self.docker_plugin.plugin_pid()
                switch_path = os.path.join(self.docker_plugin.plugin_store_path(),
                                           self._get_bridge_name(network))
                Networking.remove_interface_ns(external_link, switch_path, plugin_pid)

            self.docker_plugin.exec_by_version(vde_delete, lambda: None)

            if re.search(r"^\w+\.\d+$", external_link):
                # Only remove VLAN interfaces, physical ones cannot be removed.
                Networking.remove_interface(external_link)

    @staticmethod
    def _get_bridge_name(network: docker.models.networks.Network) -> str:
        """Return the name of the host bridge associated to the Docker network.

        Args:
            network (docker.models.networks.Network): A Docker network.

        Returns:
            str: The name of the Docker bridge in the format "kt-<network.id[:12]>".
        """
        return f"kt-{network.id[:12]}"

    @staticmethod
    def get_network_name(link: Link) -> str:
        """Return the name of a Docker network.

        Args:
            link (Kathara.model.Link): A Kathara collision domain.

        Returns:
            str: The name of the Docker network in the format "|net_prefix|_|username_prefix|_|name|".
                If shared collision domains, the format is: "|net_prefix|_|lab_hash|".
        """
        if Setting.get_instance().shared_cds == SharedCollisionDomainsOption.LABS:
            return f"{Setting.get_instance().net_prefix}_{utils.get_current_user_name()}_{link.name}"
        elif Setting.get_instance().shared_cds == SharedCollisionDomainsOption.USERS:
            return f"{Setting.get_instance().net_prefix}_{link.name}"
        elif Setting.get_instance().shared_cds == SharedCollisionDomainsOption.NOT_SHARED:
            return f"{Setting.get_instance().net_prefix}_{utils.get_current_user_name()}_{link.name}_{link.lab.hash}"



from re import match

from .base import BaseValidator


class RegexValidator(BaseValidator):

    def __init__(self, pattern):
        super(RegexValidator, self).__init__()
        self.__pattern = pattern

    @property
    def pattern(self):
        return self.__pattern

    def validate(self, input_string):
        """
        Validate input_string against a regex pattern

        :return: True if match / False otherwise
        """
        validation_result = False
        try:
            validation_result = bool(match(pattern=self.pattern, string=input_string))
        except TypeError as e:
            self.log.error(
                'Exception while validating Regex, pattern={}, input_string={} - exception: {}'.format(self.pattern,
                                                                                                       input_string,
                                                                                                       e))
        return validation_result



import platform
import subprocess
import sys
import textwrap


class Screen(object):
    """
    Class representing a console screen.
    """

    def __init__(self):
        self.__tw = textwrap.TextWrapper()
        # TODO get actual screen size
        self.__height = 40
        self.__width = 80

    @property
    def screen_height(self):
        """
        int: The screen height in rows.
        """
        return self.__height

    @property
    def screen_width(self):
        """
        int: The screen width in columns.
        """
        return self.__width

    @staticmethod
    def clear():
        """
        Clear the screen.
        """
        if platform.system() == 'Windows':
            subprocess.check_call('cls', shell=True)
        else:
            print(subprocess.check_output('clear').decode())

    @staticmethod
    def flush():
        """
        Flush any buffered standard output to screen.
        """
        sys.stdout.flush()

    def input(self, prompt=''):
        """
        Prompt the end user for input.

        Args:
            prompt (:obj:`str`, optional): The message to display as the prompt.

        Returns:
            The input provided by the user.
        """
        if sys.version[0] == '2':
            return raw_input(prompt)
        else:
            return input(prompt)

    @staticmethod
    def printf(*args):
        """
        Print the specified arguments to the screen.

        Args:
            *args: Variable length argument list.
        """
        print(*args, end='')

    @staticmethod
    def println(*args):
        """
        Print the specified arguments to the screen, including an appended newline character.

        Args:
            *args: Variable length argument list.
        """
        print(*args)



import argparse
from typing import List

from rich.live import Live

from ..ui.utils import create_lab_table
from ..ui.utils import create_panel, LabMetaHighlighter, create_topology_table
from ... import utils
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...model.Lab import Lab
from ...model.Link import BRIDGE_LINK_NAME
from ...parser.netkit.LabParser import LabParser
from ...strings import strings, wiki_description


class LinfoCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara linfo',
            description=strings['linfo'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show a help message and exit.'
        )

        self.parser.add_argument(
            '-d', '--directory',
            required=False,
            help='Specify the folder containing the network scenario.'
        )

        group = self.parser.add_mutually_exclusive_group(required=False)

        group.add_argument(
            '-w', '-l', '--watch', '--live',
            required=False,
            action='store_true',
            help='Watch mode, can be used only when a network scenario is launched.'
        )

        group.add_argument(
            '-c', '--conf',
            required=False,
            action='store_true',
            help='Read static information from lab.conf.'
        )

        topology_group = self.parser.add_mutually_exclusive_group(required=False)

        topology_group.add_argument(
            '-n', '--name',
            metavar='DEVICE_NAME',
            required=False,
            help='Show only information about a specified device.'
        )

        topology_group.add_argument(
            '-t', '--topology',
            required=False,
            action='store_true',
            help='Get running topology info'
        )

    def run(self, current_path: str, argv: List[str]) -> int:
        self.parse_args(argv)
        args = self.get_args()

        lab_path = args['directory'].replace('"', '').replace("'", '') if args['directory'] else current_path
        lab_path = utils.get_absolute_path(lab_path)

        # Load custom 'kathara.conf' if it exists
        self._load_custom_configuration(lab_path)

        try:
            lab = LabParser.parse(lab_path)
        except (Exception, IOError):
            lab = Lab(None, path=lab_path)

        if args['watch']:
            if args['name']:
                self._get_machine_live_info(lab, args['name'])
            elif args['topology']:
                self._get_topology_live_info(lab)
            else:
                self._get_lab_live_info(lab)

            return 0

        if args['conf']:
            self._get_conf_info(lab, machine_name=args['name'])

            return 0

        with self.console.status(
                f"Loading...",
                spinner="dots"
        ) as _:
            if args['name']:
                machine_stats = next(Kathara.get_instance().get_machine_stats(args['name'], lab.hash))
                message = str(machine_stats) if machine_stats else f"Device `{args['name']}` Not Found."
                style = None if machine_stats else "red bold"

                self.console.print(create_panel(message, title=f"{args['name']} Information", style=style))
            elif args['topology']:
                Kathara.get_instance().update_lab_from_api(lab)
                self.console.print(create_topology_table(lab))
            else:
                machines_stats = Kathara.get_instance().get_machines_stats(lab.hash)
                self.console.print(create_lab_table(machines_stats))

        return 0

    def _get_machine_live_info(self, lab: Lab, machine_name: str) -> None:
        with Live(None, refresh_per_second=12.5, screen=True) as live:
            live.update(self.console.status(f"Loading...", spinner="dots"))
            live.refresh_per_second = 1
            while True:
                machine_stats = next(Kathara.get_instance().get_machine_stats(machine_name, lab.hash))
                message = str(machine_stats) if machine_stats else f"Device `{machine_name}` Not Found."
                style = None if machine_stats else "red bold"

                live.update(create_panel(message, title=f"{machine_name} Information", style=style))

    def _get_topology_live_info(self, lab: Lab) -> None:
        with Live(None, refresh_per_second=1, screen=True) as live:
            live.update(self.console.status(f"Loading...", spinner="dots"))
            while True:
                Kathara.get_instance().update_lab_from_api(lab)
                table = create_topology_table(lab)
                if not table:
                    break

                live.update(table)

    def _get_lab_live_info(self, lab: Lab) -> None:
        machines_stats = Kathara.get_instance().get_machines_stats(lab.hash)

        with Live(None, refresh_per_second=12.5, screen=True) as live:
            live.update(self.console.status(f"Loading...", spinner="dots"))
            live.refresh_per_second = 1
            while True:
                table = create_lab_table(machines_stats)
                if not table:
                    break

                live.update(table)

    def _get_conf_info(self, lab: Lab, machine_name: str = None) -> None:
        if machine_name:
            self.console.print(
                create_panel(
                    str(lab.machines[machine_name]),
                    title=f"{machine_name} Information"
                )
            )
            return

        lab_meta_information = str(lab)
        if lab_meta_information:
            meta_highlighter = LabMetaHighlighter()
            self.console.print(
                create_panel(
                    meta_highlighter(lab_meta_information),
                    title="Network Scenario Information",
                )
            )

        n_machines = len(lab.machines)
        n_links = len(lab.links) if BRIDGE_LINK_NAME not in lab.links else len(lab.links) - 1

        self.console.print(
            create_panel(
                f"There are [bold green]{n_machines}[/bold green] devices.\n"
                f"There are [bold green]{n_links}[/bold green] collision domains.",
                title="Topology Information"
            )
        )



from __future__ import annotations

from typing import List, Callable, Any, Optional, Tuple, Dict

from ..exceptions import InstantiationError


class EventDispatcher(object):
    """Class implementing an event dispatcher for custom events."""
    __slots__ = ['events']

    __instance: EventDispatcher = None

    @staticmethod
    def get_instance() -> EventDispatcher:
        """Get an instance of the Dispatcher.

        Returns:
            EventDispatcher: An instance of the class.

        Raises:
            InstantiationError: If two instances of the class are created.
        """
        if EventDispatcher.__instance is None:
            EventDispatcher()

        return EventDispatcher.__instance

    def __init__(self) -> None:
        if EventDispatcher.__instance is not None:
            raise InstantiationError("This class is a singleton!")
        else:
            self.events: Dict[str, List[Tuple[Callable, Optional[Callable]]]] = {}

            EventDispatcher.__instance = self

    def get_subscribers(self, event: str) -> List[Callable]:
        """Return subscribers of a given event.

        Args:
            event (str): Name of the event.

        Returns:
            List[Callable]: Subscribed callbacks for the event.
        """
        return [run_callback for (run_callback, _) in self.events[event]]

    def register(self, event: str, obj: Any, method: Optional[str] = None) -> None:
        """Register a subscriber for a given event. Subscriber is composed of a class instance and
        an optional method name. If method name is not provided, by default the `run` method is searched.
        It also checks if the `unregister` method is present in the object, that is called during event unregistration.

        Args:
            event (str): Name of the event.
            obj (Any): instance of a class of a subscriber.
            method (Optional[str]): method name of the subscriber to run as callback. If method name is not provided,
                by default the `run` method is searched.

        Returns:
            None
        """
        if event not in self.events:
            self.events[event] = []

        self.events[event].append(
            (
                getattr(obj, 'run' if not method else method),
                getattr(obj, 'unregister') if hasattr(obj, 'unregister') else None
            )
        )

    def dispatch(self, event: str, **kwargs) -> None:
        """Dispatch a given event.

        Args:
            event (str): Name of the event.
            kwargs: Arguments to pass to the event callback.

        Returns:
            None
        """
        if event not in self.events:
            return

        for (run_callback, _) in self.events[event]:
            run_callback(**kwargs)

    def unregister(self, event: str) -> None:
        """Unregister all callbacks of a given event.

        Args:
            event (str): Name of the event.

        Returns:
            None
        """
        if event not in self.events:
            return

        for (_, unregister_callback) in self.events[event]:
            if unregister_callback:
                unregister_callback()

        del self.events[event]



import argparse
from typing import List

from .LcleanCommand import LcleanCommand
from .LstartCommand import LstartCommand
from ...foundation.cli.command.Command import Command
from ...strings import strings, wiki_description


class LrestartCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara lrestart',
            description=strings['lrestart'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show a help message and exit.'
        )

        group = self.parser.add_mutually_exclusive_group(required=False)

        group.add_argument(
            "--noterminals",
            action="store_const",
            dest="terminals",
            const=False,
            default=True,
            help='Start the network scenario without opening terminal windows.'
        )
        group.add_argument(
            "--terminals",
            action="store_const",
            dest="terminals",
            const=True,
            help='Start the network scenario opening terminal windows.'
        )
        self.parser.add_argument(
            "--privileged",
            action="store_const",
            const=True,
            required=False,
            help='Start the devices in privileged mode. MUST BE ROOT FOR THIS OPTION.'
        )
        self.parser.add_argument(
            '-d', '--directory',
            required=False,
            help='Specify the folder containing the network scenario.'
        )
        self.parser.add_argument(
            '-F', '--force-lab',
            dest='force_lab',
            required=False,
            action='store_true',
            help='Force the network scenario to start without a lab.conf or lab.dep file.'
        )
        self.parser.add_argument(
            '-l', '--list',
            required=False,
            action='store_true',
            help='Show information about running devices after the network scenario has been started.'
        )
        self.parser.add_argument(
            '-o', '--pass',
            dest='global_machine_metadata',
            metavar="METADATA",
            nargs='*',
            required=False,
            help="Apply metadata to all devices of a network scenario during startup."
        )
        self.parser.add_argument(
            '--xterm',
            required=False,
            help='Set a different terminal emulator application (Unix only).'
        )
        hosthome_group = self.parser.add_mutually_exclusive_group(required=False)
        hosthome_group.add_argument(
            '--no-hosthome', '-H',
            dest="hosthome_mount",
            action="store_const",
            const=False,
            help='Do not mount "/hosthome" directory inside devices.'
        )
        hosthome_group.add_argument(
            '--hosthome',
            dest="hosthome_mount",
            action="store_const",
            const=True,
            help='Mount "/hosthome" directory inside devices.'
        )
        shared_group = self.parser.add_mutually_exclusive_group(required=False)
        shared_group.add_argument(
            '--no-shared', '-S',
            dest="shared_mount",
            action="store_const",
            const=False,
            help='Do not mount "/shared" directory inside devices.'
        )
        shared_group.add_argument(
            '--shared',
            dest="shared_mount",
            action="store_const",
            const=True,
            help='Mount "/shared" directory inside devices.'
        )
        self.parser.add_argument(
            '--exclude',
            dest='excluded_machines',
            metavar='DEVICE_NAME',
            nargs='+',
            default=[],
            help='Exclude specified devices.'
        )
        self.parser.add_argument(
            'machine_name',
            metavar='DEVICE_NAME',
            nargs='*',
            help='Restarts only specified devices.'
        )

    def run(self, current_path: str, argv: List[str]) -> int:
        self.parse_args(argv)
        args = self.get_args()

        lclean_argv = ['-d', args['directory']] if args['directory'] else []

        if args['machine_name']:
            lclean_argv.extend(args['machine_name'])
        if args['excluded_machines']:
            lclean_argv.extend(['--exclude'] + args['excluded_machines'])

        LcleanCommand().run(current_path, lclean_argv)
        LstartCommand().run(current_path, argv)

        return 0



import hashlib
import json
import logging
import os
import re
import shlex
import signal
import sys
import tarfile
import tempfile
import threading
import uuid
from multiprocessing.dummy import Pool
from typing import Optional, Set, List, Union, Generator, Tuple, Dict, Any

import chardet
from kubernetes import client
from kubernetes.client.api import apps_v1_api
from kubernetes.client.api import core_v1_api
from kubernetes.client.rest import ApiException
from kubernetes.stream import stream
from kubernetes.watch import watch

from .KubernetesConfigMap import KubernetesConfigMap
from .KubernetesNamespace import KubernetesNamespace
from .exec_stream.KubernetesExecStream import KubernetesExecStream
from .stats.KubernetesMachineStats import KubernetesMachineStats
from ... import utils
from ...event.EventDispatcher import EventDispatcher
from ...exceptions import MachineAlreadyExistsError, MachineNotReadyError, MachineNotRunningError, MachineBinaryError, \
    InvocationError, MountDeniedError
from ...model.Lab import Lab
from ...model.Machine import Machine
from ...setting.Setting import Setting

RP_FILTER_NAMESPACE = "net.ipv4.conf.%s.rp_filter"
MAX_RESTART_COUNT = 3
MAX_TIME_ERROR = 180

OCI_RUNTIME_RE = re.compile(
    r"OCI runtime exec failed"
)

# Known commands that each container should execute
# Run order: shared.startup, machine.startup and machine.meta['exec_commands']
STARTUP_COMMANDS = [
    # If execution flag file is found, abort (this means that postStart has been called again)
    # If not flag the startup execution with a file
    "if [ -f \"/tmp/post_start\" ]; then exit; else touch /tmp/post_start; fi",

    "{sysctl_commands}",

    # Removes /etc/bind already existing configuration from k8s internal DNS
    "rm -Rf /etc/bind/*",

    # Unmount the /etc/resolv.conf and /etc/hosts files, automatically mounted by Kubernetes inside the container.
    # In this way, they can be overwritten by custom user files.
    "umount /etc/resolv.conf",
    "umount /etc/hosts",

    # Parse hostlab.b64 (if present)
    "if [ -f \"/tmp/kathara/hostlab.b64\" ]; then "
    "base64 -d /tmp/kathara/hostlab.b64 > /hostlab.tar.gz",
    # Extract hostlab.tar.gz data into /
    "tar xmfz /hostlab.tar.gz -C /; rm -f hostlab.tar.gz",
    "fi",

    # Copy the machine folder (if present) from the hostlab directory into the root folder of the container
    # In this way, files are all replaced in the container root folder
    "if [ -d \"/hostlab/{machine_name}\" ]; then "
    "(cd /hostlab/{machine_name} && tar c .) | (cd / && tar xhf - --no-same-owner --no-same-permissions); fi",

    # If /etc/hosts is not configured by the user, add the default mappings
    "if [ ! -s \"/etc/hosts\" ]; then "
    "echo '127.0.0.1 localhost' > /etc/hosts",
    "echo '::1 localhost' >> /etc/hosts",
    "echo '127.0.1.1 {machine_name}' >> /etc/hosts",
    "fi",

    # Give proper permissions to /var/www
    "if [ -d \"/var/www\" ]; then "
    "chmod -R 777 /var/www/*; fi",

    # Give proper permissions to Quagga files (if present)
    "if [ -d \"/etc/quagga\" ]; then "
    "chown -R quagga:quagga /etc/quagga/",
    "chmod 640 /etc/quagga/*; fi",

    # Give proper permissions to FRR files (if present)
    "if [ -d \"/etc/frr\" ]; then "
    "chown -R frr:frr /etc/frr/",
    "chmod 640 /etc/frr/*; fi",

    # If shared.startup file is present
    "if [ -f \"/hostlab/shared.startup\" ]; then "
    # Give execute permissions to the file and execute it
    # We redirect the output "&>" to a debugging file
    "chmod u+x /hostlab/shared.startup",
    # Adds a line to enable command output
    "sed -i \"1s;^;set -x\\n\\n;\" /hostlab/shared.startup",
    "/hostlab/shared.startup &> /var/log/shared.log; fi",

    # If .startup file is present
    "if [ -f \"/hostlab/{machine_name}.startup\" ]; then "
    # Give execute permissions to the file and execute it
    # We redirect the output "&>" to a debugging file
    "chmod u+x /hostlab/{machine_name}.startup",
    # Adds a line to enable command output
    "sed -i \"1s;^;set -x\\n\\n;\" /hostlab/{machine_name}.startup",
    "/hostlab/{machine_name}.startup &> /var/log/startup.log; fi",

    # Remove the Kubernetes' default gateway which points to the eth0 interface and causes problems sometimes.
    "ip route del default dev eth0 || true",

    # Placeholder for user commands
    "{machine_commands}",

    "touch /tmp/EOS"
]

SHUTDOWN_COMMANDS = [
    # If machine.shutdown file is present
    "if [ -f \"/hostlab/{machine_name}.shutdown\" ]; then "
    # Give execute permissions to the file and execute it
    "chmod u+x /hostlab/{machine_name}.shutdown; /hostlab/{machine_name}.shutdown; fi",

    # If shared.shutdown file is present
    "if [ -f \"/hostlab/shared.shutdown\" ]; then "
    # Give execute permissions to the file and execute it
    "chmod u+x /hostlab/shared.shutdown; /hostlab/shared.shutdown; fi"
]


class KubernetesMachine(object):
    """Class responsible for managing Kathara devices representation in Kubernetes."""
    __slots__ = ['client', 'core_client', 'kubernetes_config_map', 'kubernetes_namespace']

    def __init__(self, kubernetes_namespace: KubernetesNamespace) -> None:
        self.client: apps_v1_api.AppsV1Api = apps_v1_api.AppsV1Api()
        self.core_client: core_v1_api.CoreV1Api = core_v1_api.CoreV1Api()

        self.kubernetes_config_map: KubernetesConfigMap = KubernetesConfigMap()

        self.kubernetes_namespace: KubernetesNamespace = kubernetes_namespace

    def deploy_machines(self, lab: Lab, selected_machines: Set[str] = None, excluded_machines: Set[str] = None) -> None:
        """Deploy all the devices contained in lab.machines.

        Args:
            lab (Kathara.model.Lab.Lab): A Kathara network scenario.
            selected_machines (Set[str]): A set containing the name of the devices to deploy.
            excluded_machines (Set[str]): A set containing the name of the devices to exclude.

        Returns:
            None

        Raises:
            InvocationError: If both `selected_machines` and `excluded_machines` are specified.
        """
        if selected_machines and excluded_machines:
            raise InvocationError(f"You can either specify `selected_machines` or `excluded_machines`.")

        machines = lab.machines.items()
        if selected_machines:
            machines = {
                k: v for k, v in machines if k in selected_machines
            }.items()
        elif excluded_machines:
            machines = {
                k: v for k, v in machines if k not in excluded_machines
            }.items()

        # Do not open terminals on Megalos
        Setting.get_instance().open_terminals = False

        policy = Setting.get_instance().volume_mount_policy
        lab.add_option("_mount_volumes", policy in ['Prompt', 'Always'])
        machines_with_volumes = dict(filter(lambda x: len(x[1].meta['volumes']) > 0, machines))
        if len(machines_with_volumes) > 0:
            EventDispatcher.get_instance().dispatch(
                "machines_with_volumes", lab=lab, machines_with_volumes=machines_with_volumes
            )

        wait_thread = threading.Thread(
            target=self._wait_machines_startup,
            args=(lab, set([k for k, _ in machines]) if selected_machines or excluded_machines else None)
        )
        wait_thread.start()

        # Deploy all lab machines.
        # If there is no lab.dep file, machines can be deployed using multithreading.
        # If not, they're started sequentially
        if not lab.has_dependencies:
            pool_size = utils.get_pool_size()
            items = utils.chunk_list(machines, pool_size)

            with Pool(pool_size) as machines_pool:
                for chunk in items:
                    machines_pool.map(func=self._deploy_machine, iterable=chunk)
        else:
            for item in machines:
                self._deploy_machine(item)

        wait_thread.join()

        # Delete to avoid keeping dirty state
        del lab.general_options['_mount_volumes']

    def _wait_machines_startup(self, lab: Lab, selected_machines: Set[str]) -> None:
        """Wait the startup of the selected machines. Return when the selected machines become `Ready`.

        Args:
            lab (Lab): The network scenario of the devices to wait.
            selected_machines (Set[str]): A set containing the name of the devices to wait.

        Returns:
            None
        """
        machines = {k: v for (k, v) in lab.machines.items() if k in selected_machines}.items() if selected_machines \
            else lab.machines.items()

        EventDispatcher.get_instance().dispatch("machines_deploy_started", items=machines)

        machines_ready = 0
        machines_failed = 0

        # Create a timer to raise an exception if the execution gets stuck
        def raise_timeout_error():
            logging.error(
                f"Network scenario startup is not responding for over {MAX_TIME_ERROR} seconds, exiting. "
                f"To check devices status, use the following command:\n\t"
                f"kubectl -n {lab.hash} get pods"
            )

            # We should send a SIGINT to the main thread to exit the w.stream
            os.kill(os.getpid(), signal.SIGINT)

        timer = threading.Timer(MAX_TIME_ERROR, raise_timeout_error)
        timer.start()

        w = watch.Watch()
        for event in w.stream(self.kubernetes_namespace.client.list_namespaced_pod, namespace=lab.hash):
            # Every new event, cancel and create the timer
            timer.cancel()
            timer = threading.Timer(MAX_TIME_ERROR, raise_timeout_error)
            timer.start()

            machine_name = event['object'].metadata.labels['name']

            if not selected_machines or machine_name in selected_machines:
                logging.debug(f"Event: {event['type']} - Pod: {event['object'].metadata.name} (Device {machine_name})")

                if event['object'].status.container_statuses:
                    restart_count = event['object'].status.container_statuses[0].restart_count

                    if event['object'].status.container_statuses[0].ready:
                        machines_ready += 1
                        logging.debug(f"Device `{machine_name}` ready.")

                        EventDispatcher.get_instance().dispatch("machine_deployed", item=machine_name)
                    elif restart_count > 0:
                        if restart_count >= MAX_RESTART_COUNT:
                            logging.warning(
                                f"Stopping to wait device `{machine_name}` since it restarted more than "
                                f"{MAX_RESTART_COUNT} times. "
                                f"For a detailed log use the following command:\n\t"
                                f"kubectl -n {lab.hash} describe pod {event['object'].metadata.name}"
                            )

                            machines_failed += 1
                        elif event['object'].status.container_statuses[0].state.waiting and \
                                event['object'].status.container_statuses[0].state.waiting.reason == "CrashLoopBackOff":
                            logging.warning(f"Device `{machine_name}` has been restarted {restart_count} times.")

            if machines_ready + machines_failed == len(machines):
                # Finished watching, cancel the last timer
                timer.cancel()

                w.stop()

        if machines_ready == len(machines):
            EventDispatcher.get_instance().dispatch("machines_deploy_ended")

    def _deploy_machine(self, machine_item: Tuple[str, Machine]) -> None:
        """Deploy a Kubernetes deployment from the Kathara device contained in machine_item.

        Args:
           machine_item (Tuple[str, Machine]): A tuple composed by the name of the device and a device object

        Returns:
           None
        """
        (_, machine) = machine_item

        self.create(machine)

    def create(self, machine: Machine) -> None:
        """Create a Kubernetes deployment and a Pod representing the device and assign it to machine.api_object.

        Args:
            machine (Kathara.model.Machine.Machine): a Kathara device.

        Returns:
            None

        Raises:
            MachineAlreadyExistsError: If a device with the name specified already exists.
        """
        logging.debug("Creating device `%s`..." % machine.name)

        # Get the global machine metadata into a local variable (just to avoid accessing the lab object every time)
        global_machine_metadata = machine.lab.global_machine_metadata

        # If privileged is defined for the device, throw a warning.
        if "privileged" in global_machine_metadata or machine.is_privileged():
            logging.warning(
                f'Privileged option is not supported on Megalos. It will be ignored on device `{machine.name}`.'
            )

        # If bridged is defined for the device, throw a warning.
        if "bridged" in global_machine_metadata or machine.is_bridged():
            logging.warning(
                f'Bridged option is not supported on Megalos. It will be ignored on device `{machine.name}`.'
            )

        # If any ulimit is defined fot the device throw a warning
        if "ulimits" in machine.meta and len(machine.meta["ulimits"].keys()) > 0:
            logging.warning(
                f'Ulimit option is not supported on Megalos. It will be ignored on device `{machine.name}`.'
            )

        # If any exec command is passed in command line, add it.
        if "exec" in global_machine_metadata:
            machine.add_meta("exec", global_machine_metadata["exec"])

        # Sysctl params to pass to the container creation
        sysctl_parameters = {RP_FILTER_NAMESPACE % x: 0 for x in ["all", "default", "lo"]}

        sysctl_parameters["net.ipv4.ip_forward"] = 1
        sysctl_parameters["net.ipv4.icmp_ratelimit"] = 0

        if machine.is_ipv6_enabled():
            sysctl_parameters["net.ipv6.conf.all.forwarding"] = 1
            sysctl_parameters["net.ipv6.conf.all.accept_ra"] = 0
            sysctl_parameters["net.ipv6.icmp.ratelimit"] = 0
            sysctl_parameters["net.ipv6.conf.default.disable_ipv6"] = 0
            sysctl_parameters["net.ipv6.conf.all.disable_ipv6"] = 0
        else:
            sysctl_parameters["net.ipv6.conf.default.disable_ipv6"] = 1
            sysctl_parameters["net.ipv6.conf.all.disable_ipv6"] = 1
            sysctl_parameters["net.ipv6.conf.default.forwarding"] = 0
            sysctl_parameters["net.ipv6.conf.all.forwarding"] = 0

        # Merge machine sysctls
        machine.meta['sysctls'] = {**sysctl_parameters, **machine.meta['sysctls']}

        machine.add_meta('real_name', self.get_deployment_name(machine.name))

        try:
            config_map = self.kubernetes_config_map.deploy_for_machine(machine)
            machine_definition = self._build_definition(machine, config_map)

            machine.api_object = self.client.create_namespaced_deployment(body=machine_definition,
                                                                          namespace=machine.lab.hash
                                                                          )
        except ApiException as e:
            if e.status == 409 and 'Conflict' in e.reason:
                raise MachineAlreadyExistsError(machine.name)
            else:
                raise e

    def _build_definition(self, machine: Machine, config_map: client.V1ConfigMap) -> client.V1Deployment:
        """Return a Kubernetes deployment from a Kathara device and a Kubernetes ConfigMap.

        Args:
            machine (Kathara.model.Machine.Machine): A Kathara device.
            config_map (client.V1ConfigMap): A Kubernetes ConfigMap containing the tar data to upload on the deployment.

        Returns:
            client.V1Deployment: A Kubernetes deployment.
        """
        volume_mounts = []
        if config_map:
            # Define volume mounts for hostlab if a ConfigMap is defined.
            volume_mounts.append(client.V1VolumeMount(name="hostlab", mount_path="/tmp/kathara"))

        if Setting.get_instance().host_shared:
            volume_mounts.append(client.V1VolumeMount(name="shared", mount_path="/shared"))

        # Mount volumes defined in the device
        try:
            for idx, (host_path, volume) in enumerate(machine.get_volumes().items()):
                missing_permissions = utils.check_directory_permissions(host_path, volume['mode'])

                if not missing_permissions:
                    volume_mounts.append(
                        client.V1VolumeMount(
                            name=f"volume{idx}", mount_path=volume['guest_path'], read_only=volume['mode'] == "ro"
                        )
                    )
                else:
                    raise PermissionError(
                        f"To mount volume `{host_path}` in `{volume['guest_path']}` "
                        f"you miss the following permissions: `{', '.join(missing_permissions)}`."
                    )
        except MountDeniedError:
            logging.warning(f"Volumes of device `{machine.name}` will not be mounted.")

        # Machine must be executed in privileged mode to run sysctls.
        security_context = client.V1SecurityContext(privileged=True)

        ports_info = machine.get_ports()
        container_ports = None
        if ports_info:
            container_ports = []
            for (host_port, protocol), guest_port in ports_info.items():
                container_ports.append(
                    client.V1ContainerPort(
                        name=str(uuid.uuid4()).replace('-', '')[0:15],
                        container_port=guest_port,
                        host_port=host_port,
                        protocol=protocol.upper()
                    )
                )

        resources = None
        memory = machine.get_mem()
        cpus = machine.get_cpu(multiplier=1000)
        if memory or cpus:
            limits = dict()
            if memory:
                limits["memory"] = memory.upper()
            if cpus:
                limits["cpu"] = "%dm" % cpus

            resources = client.V1ResourceRequirements(limits=limits)

        shell = machine.meta["shell"] if "shell" in machine.meta else Setting.get_instance().device_shell

        # postStart lifecycle hook is launched asynchronously by k8s master when the main container is Ready
        # On Ready state, the pod has volumes and network interfaces up, so this hook is used
        # to execute custom commands coming from .startup file and "exec" option
        # Build the final startup commands string
        sysctl_commands = "; ".join(["sysctl -w -q %s=%d" % item for item in machine.meta["sysctls"].items()])
        machine_commands = "; ".join(machine.meta['exec_commands']) if machine.meta['exec_commands'] else ":"

        startup_commands_string = "; ".join(STARTUP_COMMANDS) \
            .format(machine_name=machine.name, sysctl_commands=sysctl_commands, machine_commands=machine_commands)

        post_start = client.V1LifecycleHandler(
            _exec=client.V1ExecAction(
                command=[shell, "-c", startup_commands_string]
            )
        )
        lifecycle = client.V1Lifecycle(post_start=post_start)

        env = [
            client.V1EnvVar("_MEGALOS_SHELL", shell)
        ]
        for env_var_name, env_var_value in machine.meta['envs'].items():
            env.append(client.V1EnvVar(env_var_name, env_var_value))

        entrypoint = shlex.split(machine.meta["entrypoint"]) if "entrypoint" in machine.meta else None
        args = machine.meta["args"] if "args" in machine.meta and machine.meta["args"] else None
        if args:
            args = shlex.split(args) if type(args) == str else args

        container_definition = client.V1Container(
            name=machine.meta['real_name'],
            image=machine.get_image(),
            lifecycle=lifecycle,
            stdin=True,
            tty=True,
            image_pull_policy=Setting.get_instance().image_pull_policy,
            ports=container_ports,
            resources=resources,
            volume_mounts=volume_mounts,
            security_context=security_context,
            env=env,
            command=entrypoint,
            args=args
        )

        pod_annotations = {}
        network_interfaces = []
        for (idx, interface) in machine.interfaces.items():
            additional_data = {"kathara.link": interface.link.name}
            if interface.mac_address:
                additional_data["mac"] = interface.mac_address

            network_interfaces.append({
                "name": interface.link.api_object["metadata"]["name"],
                "namespace": machine.lab.hash,
                "interface": "net%d" % idx,
                **additional_data
            })
        pod_annotations["k8s.v1.cni.cncf.io/networks"] = json.dumps(network_interfaces)

        # Create labels (so Deployment can match them)
        pod_labels = {"name": machine.name,
                      "app": "kathara"
                      }

        pod_metadata = client.V1ObjectMeta(deletion_grace_period_seconds=0,
                                           annotations=pod_annotations,
                                           labels=pod_labels
                                           )

        # Add fake DNS just to override k8s one
        dns_config = client.V1PodDNSConfig(nameservers=["127.0.0.1"])

        volumes = []
        if config_map:
            # Hostlab is the lab base64 encoded .tar.gz of the machine files, deployed as a ConfigMap in the cluster
            # The base64 file is mounted into /tmp and it's extracted by the postStart hook
            volumes.append(client.V1Volume(
                name="hostlab",
                config_map=client.V1ConfigMapVolumeSource(
                    name=config_map.metadata.name
                )
            ))

        # Container /shared mounts in /home/shared folder
        if Setting.get_instance().host_shared:
            volumes.append(client.V1Volume(
                name="shared",
                host_path=client.V1HostPathVolumeSource(
                    path='/home/shared',
                    type='DirectoryOrCreate'
                )
            ))

        for idx, (host_path, volume) in enumerate(machine.meta["volumes"].items()):
            volumes.append(client.V1Volume(
                name=f"volume{idx}",
                host_path=client.V1HostPathVolumeSource(
                    path=host_path,
                    type='DirectoryOrCreate'
                )
            ))

        # Add Docker Config JSON for Private Registries
        image_pull_secrets = []
        if Setting.get_instance().docker_config_json:
            image_pull_secrets.append(client.V1LocalObjectReference(name='private-registry'))

        pod_spec = client.V1PodSpec(
            containers=[container_definition],
            hostname=machine.meta['real_name'],
            dns_policy="None",
            dns_config=dns_config,
            volumes=volumes,
            image_pull_secrets=image_pull_secrets
        )

        pod_template = client.V1PodTemplateSpec(metadata=pod_metadata, spec=pod_spec)
        selector_rules = client.V1LabelSelector(match_labels=pod_labels)
        deployment_spec = client.V1DeploymentSpec(replicas=1,
                                                  template=pod_template,
                                                  selector=selector_rules
                                                  )
        deployment_metadata = client.V1ObjectMeta(name=machine.meta['real_name'], labels=pod_labels)

        return client.V1Deployment(api_version="apps/v1",
                                   kind="Deployment",
                                   metadata=deployment_metadata,
                                   spec=deployment_spec
                                   )

    def undeploy(self, lab_hash: str, selected_machines: Set[str] = None, excluded_machines: Set[str] = None) -> None:
        """Undeploy all the running Kubernetes deployments and Pods contained in the scenario defined by the lab_hash.

        If selected_machines is not None, undeploy only the specified devices.

        Args:
            lab_hash (str): The hash of the network scenario to undeploy.
            selected_machines (Set[str]): A set containing the name of the devices to undeploy.
            excluded_machines (Set[str]): A set containing the name of the devices to exclude.

        Returns:
            None

        Raises:
            InvocationError: If both `selected_machines` and `excluded_machines` are specified.
        """
        if selected_machines and excluded_machines:
            raise InvocationError(f"You can either specify `selected_machines` or `excluded_machines`.")

        pods = self.get_machines_api_objects_by_filters(lab_hash=lab_hash)
        machines_to_watch = {item.metadata.labels["name"] for item in pods}
        if selected_machines:
            pods = [item for item in pods if item.metadata.labels["name"] in selected_machines]
            machines_to_watch = selected_machines if selected_machines else machines_to_watch
        elif excluded_machines:
            pods = [item for item in pods if item.metadata.labels["name"] not in excluded_machines]
            machines_to_watch = machines_to_watch if not excluded_machines else machines_to_watch - excluded_machines

        if len(pods) > 0:
            wait_thread = threading.Thread(target=self._wait_machines_shutdown, args=(lab_hash, machines_to_watch))
            wait_thread.start()

            pool_size = utils.get_pool_size()
            items = utils.chunk_list(pods, pool_size)

            with Pool(pool_size) as machines_pool:
                for chunk in items:
                    machines_pool.map(func=self._undeploy_machine, iterable=chunk)

            wait_thread.join()

    def _wait_machines_shutdown(self, lab_hash: str, selected_machines: Set[str]):
        """Wait the shutdown of the selected machines. Return when all the selected machines are terminated.

        Args:
            lab_hash (str): The hash of the network scenario of the devices to wait.
            selected_machines (Set[str]): A set containing the name of the devices to wait.

        Returns:
            None
        """
        EventDispatcher.get_instance().dispatch("machines_undeploy_started", items=selected_machines)

        machines_cleaned = 0

        w = watch.Watch()
        for event in w.stream(self.kubernetes_namespace.client.list_namespaced_pod, namespace=lab_hash):
            machine_name = event['object'].metadata.labels['name']

            if machine_name in selected_machines:
                logging.debug(f"Event: {event['type']} - Pod: {event['object'].metadata.name} (Device {machine_name})")

                if event['type'] == "DELETED":
                    EventDispatcher.get_instance().dispatch("machine_undeployed", item=machine_name)

                    machines_cleaned += 1

            if machines_cleaned == len(selected_machines):
                EventDispatcher.get_instance().dispatch("machines_undeploy_ended")
                w.stop()

    def wipe(self) -> None:
        """Undeploy all the running Kubernetes deployments and Pods.

        Returns:
            None
        """
        pods = self.get_machines_api_objects_by_filters()

        pool_size = utils.get_pool_size()
        items = utils.chunk_list(pods, pool_size)

        with Pool(pool_size) as machines_pool:
            for chunk in items:
                machines_pool.map(func=self._undeploy_machine, iterable=chunk)

    def _undeploy_machine(self, pod_api_object: client.V1Pod) -> None:
        """Undeploy a Kubernetes pod.

        Args:
            pod_api_object (client.V1Pod): The Kubernetes pod to undeploy.

        Returns:
            None
        """

        self._delete_machine(pod_api_object)

    def _delete_machine(self, pod_api_object: client.V1Pod) -> None:
        """Delete the Kubernetes deployment and Pod associated to pod_api_object.

        Args:
            pod_api_object (client.V1Pod): A Kubernetes Pod API object.

        Returns:
            None
        """
        machine_name = pod_api_object.metadata.labels["name"]
        machine_namespace = pod_api_object.metadata.namespace

        # Build the shutdown command string
        shutdown_commands_string = "; ".join(SHUTDOWN_COMMANDS).format(machine_name=machine_name)

        try:
            shell_env_value = self.get_env_var_value_from_pod(pod_api_object, "_MEGALOS_SHELL")
            shell = shell_env_value if shell_env_value else Setting.get_instance().device_shell
            self.exec(machine_namespace, machine_name, command=[shell, '-c', shutdown_commands_string], is_stream=False)
        except ApiException:
            pass
        except MachineNotRunningError:
            pass

        deployment_name = self.get_deployment_name(machine_name)
        self.kubernetes_config_map.delete_for_machine(deployment_name, machine_namespace)
        self.client.delete_namespaced_deployment(name=deployment_name, namespace=machine_namespace)

    def connect(self, lab_hash: str, machine_name: str, shell: Union[str, List[str]] = None, logs: bool = False) \
            -> None:
        """Open a stream to the Kubernetes Pod specified by machine_name using the specified shell.

        Args:
            lab_hash (str): The hash of the network scenario containing the device.
            machine_name (str): The name of the device to connect.
            shell (Union[str, List[str]]): The path to the desired shell.
            logs (bool): If True, print the logs of the startup command.

        Returns:
            None

        Raises:
            MachineNotRunningError: If the specified device is not running.
            MachineNotReadyError: If the device is not ready.
        """
        pods = self.get_machines_api_objects_by_filters(lab_hash=lab_hash, machine_name=machine_name)
        if not pods:
            raise MachineNotRunningError(machine_name)
        deployment = pods.pop()

        if 'Running' not in deployment.status.phase:
            raise MachineNotReadyError(machine_name)

        if not shell:
            shell_env_value = self.get_env_var_value_from_pod(deployment, "_MEGALOS_SHELL")
            shell = shlex.split(shell_env_value if shell_env_value else Setting.get_instance().device_shell)
        else:
            shell = shlex.split(shell) if type(shell) is str else shell

        logging.debug("Connect to device `%s` with shell: %s" % (machine_name, shell))

        if logs and Setting.get_instance().print_startup_log:
            exec_output = self.exec(lab_hash,
                                    machine_name,
                                    command="/bin/cat /var/log/shared.log /var/log/startup.log /var/kathara/*"
                                    )
            try:
                print("--- Startup Commands Log\n")
                while True:
                    (stdout, _) = next(exec_output)

                    char_encoding = chardet.detect(stdout) if stdout else None

                    stdout = stdout.decode(char_encoding['encoding']) if stdout else ""
                    sys.stdout.write(stdout)
            except StopIteration:
                print("\n--- End Startup Commands Log\n")
                pass

        resp = stream(self.core_client.connect_get_namespaced_pod_exec,
                      name=deployment.metadata.name,
                      namespace=lab_hash,
                      command=shell,
                      stdout=True,
                      stderr=True,
                      stdin=True,
                      tty=True,
                      _preload_content=False
                      )

        from .terminal.KubernetesWSTerminal import KubernetesWSTerminal
        KubernetesWSTerminal(resp).start()

    @staticmethod
    def get_env_var_value_from_pod(pod: client.V1Pod, var_name: str) -> Optional[str]:
        """Return the value of an environment variable of the Kubernetes Pod.

        Args:
            pod (client.V1Pod): A Kubernetes Pod.
            var_name (str): The name of the environment variable.

        Returns:
            Optional[str]: The value of the environment variable.
        """
        containers = pod.spec.containers
        if not containers:
            return None

        # There is only one container definition in Megalos Pods
        container_definition = containers.pop()
        # Get env vars list
        env_vars = container_definition.env

        if not env_vars:
            return None

        # Iterate over the env vars and search the desired one
        for env_var in env_vars:
            if var_name == env_var.name:
                return env_var.value

        return None

    def exec(self, lab_hash: str, machine_name: str, command: Union[str, List], tty: bool = False, stdin: bool = False,
             stdin_buffer: List[Union[str, bytes]] = None, stderr: bool = False, is_stream: bool = True) \
            -> Union[KubernetesExecStream, Tuple[bytes, bytes, int]]:
        """Execute the command on the Kubernetes Pod specified by the lab_hash and the machine_name.

        Args:
            lab_hash (str): The hash of the network scenario containing the device.
            machine_name (str): The name of the device.
            command (Union[str, List]): The command to execute.
            tty (bool): If True, open a new tty.
            stdin (bool): If True, open the stdin channel.
            stdin_buffer (List[Union[str, bytes]]): List of command to pass to the stdin.
            stderr (bool): If True, return the stderr.
            is_stream (bool): If True, return a KubernetesExecStream object.
                If False, returns a tuple containing the complete stdout, the stderr, and the return code of the command.

        Returns:
             Union[KubernetesExecStream, Tuple[bytes, bytes, int]]: A KubernetesExecStream object or
             a tuple containing the stdout, the stderr and the return code of the command.

        Raises:
            MachineNotRunningError: If the specified device is not running.
            MachineBinaryError: If the command specified is not found on the device.
        """
        command = shlex.split(command) if type(command) is str else command

        logging.debug("Executing command `%s` to device with name: %s" % (command, machine_name))

        try:
            # Retrieve the pod of current Deployment
            pods = self.get_machines_api_objects_by_filters(lab_hash=lab_hash, machine_name=machine_name)
            if not pods:
                raise MachineNotRunningError(machine_name)
            pod = pods.pop()

            response = stream(self.core_client.connect_get_namespaced_pod_exec,
                              name=pod.metadata.name,
                              namespace=lab_hash,
                              command=command,
                              stdout=True,
                              stderr=stderr,
                              stdin=stdin,
                              tty=tty,
                              _preload_content=False
                              )
        except ApiException as e:
            raise e

        if stdin_buffer is None:
            stdin_buffer = []

        if is_stream:
            exec_result = self._exec_stream(response, stdin, stdin_buffer, stderr)
            return KubernetesExecStream(exec_result, response)
        else:
            return self._exec_all(response, machine_name, command, stdin, stdin_buffer, stderr)

    @staticmethod
    def _exec_stream(response: Any, stdin: bool = False, stdin_buffer: List[Union[str, bytes]] = None,
                     has_stderr: bool = False) -> Generator[Tuple[bytes, bytes], None, None]:
        """Execute the command on the Kubernetes Pod, returning a generator.

        Args:
            response (Any): The stream response from Kubernetes API.
            stdin (bool): If True, open the stdin channel.
            stdin_buffer (List[Union[str, bytes]]): List of command to pass to the stdin.
            has_stderr (bool): If True, return the stderr.

        Returns:
            Generator[Tuple[bytes, bytes]]: A generator of tuples containing the stdout and stderr in bytes.
        """
        while response.is_open():
            stdout = None
            stderr = None
            if response.peek_stdout():
                stdout = response.read_stdout()
            if has_stderr and response.peek_stderr():
                stderr = response.read_stderr()
            if stdin and stdin_buffer:
                param = stdin_buffer.pop(0)
                response.write_stdin(param)
                if len(stdin_buffer) <= 0:
                    break

            yield stdout.encode('utf-8') if stdout else None, stderr.encode('utf-8') if stderr else None

        response.close()

    @staticmethod
    def _exec_all(response: Any, machine_name: str, command: List, stdin: bool = False,
                  stdin_buffer: List[Union[str, bytes]] = None, has_stderr: bool = False) -> Tuple[bytes, bytes, int]:
        """Execute the command on the Kubernetes Pod, returning the full output.

        Args:
            response (Any): The stream response from Kubernetes API.
            machine_name (str): The name of the device.
            command (List): The command to execute.
            stdin (bool): If True, open the stdin channel.
            stdin_buffer (List[Union[str, bytes]]): List of command to pass to the stdin.
            has_stderr (bool): If True, return the stderr.

        Returns:
            Tuple[bytes, bytes, int]: A tuple containing the stdout, the stderr and the return code of the command.

        Raises:
            MachineBinaryError: If the command specified is not found on the device.
        """
        result = {'stdout': '', 'stderr': ''}

        while response.is_open():
            if response.peek_stdout():
                result['stdout'] += response.read_stdout()
            if has_stderr and response.peek_stderr():
                result['stderr'] += response.read_stderr()
            if stdin and stdin_buffer:
                param = stdin_buffer.pop(0)
                response.write_stdin(param)
                if len(stdin_buffer) <= 0:
                    break

        response.close()

        try:
            exit_code = response.returncode
        except ValueError as e:
            if OCI_RUNTIME_RE.search(str(e)):
                raise MachineBinaryError(shlex.join(command), machine_name)
            exit_code = 1

        return result['stdout'].encode('utf-8'), result['stderr'].encode('utf-8'), exit_code

    def copy_files(self, machine_api_object: client.V1Deployment, path: str, tar_data: bytes) -> None:
        """Copy the files contained in tar_data in the Kubernetes deployment path specified by the machine_api_object.

        Args:
            machine_api_object (client.V1Deployment): A Kubernetes deployment.
            path (str): The path of where copy the tar_data.
            tar_data (bytes): The data to copy in the deployment.

        Returns:
            None
        """
        machine_name = machine_api_object.metadata.labels["name"]
        machine_namespace = machine_api_object.metadata.namespace

        exec_output = self.exec(machine_namespace,
                                machine_name,
                                command=['tar', 'xvfz', '-', '-C', path],
                                stdin=True,
                                stdin_buffer=[tar_data]
                                )

        try:
            next(exec_output)
        except StopIteration:
            pass

    def retrieve_files(self, machine_api_object: client.V1Deployment, src: str, dst: str) -> None:
        """Get the file or directory from the Docker container specified by the machine_api_object into dst.

        Args:
            machine_api_object (docker.models.containers.Container): A Docker container.
            src (str): The path of the file or folder to copy from the device.
            dst (str): The destination path on the host.

        Returns:
            None
        """
        machine_name = machine_api_object.metadata.labels["name"]
        machine_namespace = machine_api_object.metadata.namespace

        # Get the tar from the Pod on the stdout
        exec_output = self.exec(machine_namespace, machine_name, command=['tar', 'cf', '-', src], is_stream=True)

        # Create a tmp tar file and write the output from the Pod
        with tempfile.NamedTemporaryFile(mode='wb+', suffix='.tar') as temp_file:
            # Write chunks in the temp tar file
            while True:
                try:
                    stdout, _ = next(exec_output)
                    if stdout:
                        temp_file.write(stdout)
                except StopIteration:
                    break

            # After writing, go at the beginning of file
            temp_file.seek(0)

            # Now, we need to extract the tmp tar file into the destination folder
            with tarfile.open(fileobj=temp_file, mode='r') as tar_file:
                tar_file.extractall(path=dst)

    def get_machines_api_objects_by_filters(self, lab_hash: str = None, machine_name: str = None) -> List[client.V1Pod]:
        """Return the List of Kubernetes Pods.

        Args:
            lab_hash (str): The hash of a network scenario. If specified, return all the Kubernetes Pod in the scenario.
            machine_name (str): The name of a device. If specified, return the specified Kubernetes Pod of the scenario.

        Returns:
            List[client.V1Pod]: A list of Kubernetes Pods objects.
        """
        filters = ["app=kathara"]
        if machine_name:
            filters.append(f"name={machine_name}")

        # Get all Kathara namespaces if lab_hash is None
        namespaces = list(map(lambda x: x.metadata.name, self.kubernetes_namespace.get_all())) \
            if not lab_hash else [lab_hash]

        machines = []
        for namespace in namespaces:
            machines.extend(self.core_client.list_namespaced_pod(namespace=namespace,
                                                                 label_selector=",".join(filters),
                                                                 timeout_seconds=9999
                                                                 ).items
                            )

        return machines

    def get_machines_stats(self, lab_hash: str = None, machine_name: str = None) -> \
            Generator[Dict[str, KubernetesMachineStats], None, None]:
        """Return a generator containing the Kubernetes devices' stats.

        Args:
            lab_hash (str): The hash of a network scenario. If specified, return all the stats of the devices in the
                scenario.
            machine_name (str): The name of a device. If specified, return the specified device stats.

        Returns:
            Generator[Dict[str, KubernetesMachineStats], None, None]: A generator containing device name as keys and
                KubernetesMachineStats as values.
        """
        machines_stats = {}

        def load_machine_stats(pod):
            if pod.metadata.name not in machines_stats:
                machines_stats[pod.metadata.name] = KubernetesMachineStats(pod)

        while True:
            pods = self.get_machines_api_objects_by_filters(lab_hash=lab_hash, machine_name=machine_name)
            if not pods:
                yield dict()

            pool_size = utils.get_pool_size()
            items = utils.chunk_list(pods, pool_size)
            with Pool(pool_size) as machines_pool:
                for chunk in items:
                    machines_pool.map(func=load_machine_stats, iterable=chunk)

            machines_to_remove = []
            for machine_id, machine_stats in machines_stats.items():
                try:
                    machine_stats.update()
                except StopIteration:
                    machines_to_remove.append(machine_id)
                    continue

            for k in machines_to_remove:
                machines_stats.pop(k, None)

            yield machines_stats

    @staticmethod
    def get_deployment_name(name: str) -> str:
        """Return the name of the Kubernetes deployment corresponding to 'name'.

        Args:
            name (str): The name of a Kathara device.

        Returns:
            str: The name for the Kubernetes deployment.
        """
        suffix = ''
        # Underscore is replaced with -, but to keep name uniqueness append 8 chars of hash from the original name
        if '_' in name:
            suffix = '-%s' % hashlib.md5(name.encode('utf-8', errors='ignore')).hexdigest()[:8]
            name = name.replace('_', '-')

        machine_name = "%s-%s%s" % (Setting.get_instance().device_prefix, name, suffix)
        return re.sub(r'[^0-9a-z\-.]+', '', machine_name.lower())



import argparse
import sys
from typing import List

from ..ui.utils import confirmation_prompt
from ... import utils
from ...exceptions import PrivilegeError
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...setting.Setting import Setting
from ...strings import strings, wiki_description


class WipeCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara wipe',
            description=strings['wipe'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show a help message and exit.'
        )

        self.parser.add_argument(
            '-f', '--force',
            required=False,
            action='store_true',
            help='Force the wipe.'
        )

        group = self.parser.add_mutually_exclusive_group(required=False)

        group.add_argument(
            '-s', '--settings',
            required=False,
            action='store_true',
            help='Wipe the stored settings of the current user.'
        )

        group.add_argument(
            '-a', '--all',
            required=False,
            action='store_true',
            help='Wipe all Kathara devices and collision domains of all users. MUST BE ROOT FOR THIS OPTION.'
        )

    def run(self, current_path: str, argv: List[str]) -> int:
        self.parse_args(argv)
        args = self.get_args()

        if not args['force']:
            confirmation_prompt("Are you sure to wipe Kathara?", lambda: None, sys.exit)

        if args['settings']:
            Setting.wipe_from_disk()
        else:
            if args['all'] and not utils.is_admin():
                raise PrivilegeError("You must be root in order to wipe all Kathara devices of all users.")

            Kathara.get_instance().wipe(all_users=bool(args['all']))

        return 0



try:
    # python2
    from urlparse import urlparse
except Exception:
    # python3
    from urllib.parse import urlparse

from .base import BaseValidator


class UrlValidator(BaseValidator):

    def __init__(self):
        """
        URL Validator class
        """
        super(UrlValidator, self).__init__()

    def validate(self, input_string):
        """
        Validate url

        :return: True if match / False otherwise
        """
        parsed_url = urlparse(url=input_string)
        return bool(parsed_url.scheme and parsed_url.netloc)



from rich.console import Console
from rich.table import Table

strings = {
    "vstart": "Start a new Kathara device",
    "vclean": "Stop a single Kathara device",
    "vconfig": "Manage the network interfaces of a running Kathara device",
    "lstart": "Start a Kathara network scenario",
    "lclean": "Stop a Kathara network scenario",
    "linfo": "Show information about a Kathara network scenario",
    "lrestart": "Restart a Kathara network scenario",
    "lconfig": "Manage the network interfaces of a running Kathara device in a Kathara network scenario",
    "connect": "Connect to a Kathara device",
    "exec": "Execute a command in a Kathara device",
    "wipe": "Delete all Kathara devices and collision domains, optionally also delete settings",
    "list": "Show all running Kathara devices of the current user",
    "settings": "Show and edit Kathara settings",
    "check": "Check your system environment"
}

wiki_description = "For examples and further information visit: https://github.com/KatharaFramework/Kathara/wiki"


def formatted_strings() -> str:
    console = Console(record=True)
    commands_table = Table(show_header=False, show_edge=False, show_lines=False, box=None)
    for item in strings.items():
        commands_table.add_row(*item)

    with console.capture() as capture:
        console.print(commands_table)
    return capture.get()



from typing import Tuple, Any

from ...exceptions import ClassNotFoundError
from ...utils import class_for_name


class Factory(object):
    __slots__ = ['module_template', 'name_template']

    def get_module_name(self, args: Tuple) -> str:
        return self.module_template % args

    def get_class_name(self, args: Tuple) -> str:
        return self.name_template % args

    def get_class(self, module_args: Tuple = (), class_args: Tuple = ()) -> Any:
        module_name = self.get_module_name(module_args)
        class_name = self.get_class_name(class_args)

        try:
            return class_for_name(module_name, class_name)
        except ImportError as e:
            if e.name == "%s.%s" % (module_name, class_name):
                raise ClassNotFoundError
            else:
                raise ImportError from e

    def create_instance(self, module_args: Tuple = (), class_args: Tuple = ()) -> Any:
        return self.get_class(module_args, class_args)()



import json
import os.path

from ..trdparty.consolemenu.validators.base import BaseValidator


class DockerConfigJsonValidator(BaseValidator):
    def __init__(self) -> None:
        super(DockerConfigJsonValidator, self).__init__()

    def validate(self, input_string: str) -> bool:
        try:
            with open(os.path.expanduser(input_string), 'r') as docker_config_json_file:
                json.load(docker_config_json_file)
            return True
        except (OSError, ValueError) as e:
            print(str(e))
            return False



import logging
from typing import Optional, Dict, List

from kubernetes import client, watch
from kubernetes.client.api import core_v1_api
from kubernetes.client.rest import ApiException

from ...model.Lab import Lab
from ...setting.Setting import Setting


class KubernetesSecret(object):
    """Class responsible for interacting with Kubernetes secrets."""
    __slots__ = ['client']

    def __init__(self) -> None:
        self.client: core_v1_api.CoreV1Api = core_v1_api.CoreV1Api()

    def create(self, lab: Lab) -> List[client.V1Secret]:
        """Create the Kubernetes secrets for a Kathara network scenario.

        Args:
            lab (Kathara.model.Lab.Lab): A Kathara network scenario.

        Returns:
            List[client.V1Secret]: A list of created Kubernetes secrets for the network scenario.
        """
        secrets = []

        docker_config_json = Setting.get_instance().docker_config_json
        if docker_config_json:
            secret = self._create_secret(
                lab.hash, "private-registry",
                "kubernetes.io/dockerconfigjson", {".dockerconfigjson": docker_config_json}
            )

            if secret:
                secrets.append(secret)

        return secrets

    def _create_secret(self, lab_hash: str, name: str, secret_type: str, data: Dict) -> Optional[client.V1Secret]:
        """Create a Kubernetes secret of a secret type for a Kathara network scenario.

        Args:
            lab_hash (str): The hash of a Kathara network scenario.
            name (str): The name of the secret.
            secret_type (str): The type of the secret.
            data (Dict): A dictionary containing the data of the secret.

        Returns:
            Optional[client.V1Secret]: The Kubernetes secret.
        """
        secret_definition = client.V1Secret(
            metadata=client.V1ObjectMeta(name=name, namespace=lab_hash, labels={'app': 'kathara'}),
            type=secret_type,
            data=data
        )

        try:
            self.client.create_namespaced_secret(lab_hash, secret_definition)
            self._wait_secret_creation(lab_hash, name)

            return secret_definition
        except ApiException:
            return None

    def _wait_secret_creation(self, lab_hash: str, name: str) -> None:
        """Wait for the secret creation to be completed.

        Args:
            lab_hash (str): The hash of a Kathara network scenario.
            name (str): The name of the secret.

        Returns:
            None
        """
        w = watch.Watch()
        for event in w.stream(self.client.list_namespaced_secret, namespace=lab_hash,
                              field_selector=f"metadata.name={name}"):
            logging.debug(f"Event: {event['type']} - Secret: {event['object'].metadata.name}")

            if event['type'] == "ADDED":
                break



import re
from typing import Optional

from . import Link as LinkPackage
from . import Machine as MachinePackage
from ..exceptions import InterfaceMacAddressError

MAC_ADDRESS_REGEX = re.compile(r"^([0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2}$")


class Interface(object):
    """Interface object associated to a Machine network interface.

    Attributes:
        machine (Kathara.model.Machine.Machine): The machine associated to this interface.
        link (Kathara.model.Link.Link): The collision domain associated to this interface.
        num (int): The interface number.
        mac_address (Optional[str]): The MAC address of the interface. If None, a generated MAC address
            is associated when the Machine is started.
    """
    __slots__ = ['machine', 'link', 'num', 'mac_address']

    def __init__(self, machine: 'MachinePackage.Machine', link: 'LinkPackage.Link',
                 num: int, mac_address: Optional[str] = None) -> None:
        self.machine: 'MachinePackage.Machine' = machine
        self.link: 'LinkPackage.Link' = link
        self.num: int = num
        self.mac_address: Optional[str] = mac_address

        if self.mac_address and not MAC_ADDRESS_REGEX.match(self.mac_address):
            raise InterfaceMacAddressError(self.mac_address, self.num, machine.name)

    def __repr__(self) -> str:
        return "Interface(%s, %d, %s)" % (self.machine.name, self.num, self.mac_address)



# Generic Exceptions
class ClassNotFoundError(Exception):
    pass


class HTTPConnectionError(Exception):
    pass


class InstantiationError(Exception):
    pass


class InvocationError(Exception):
    pass


# Settings Exceptions
class SettingsError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(f"Settings file is not valid: {message} Fix it or delete it before launching.")


class SettingsNotFoundError(Exception):
    def __init__(self, path: str) -> None:
        super().__init__(f"Settings file not found in path `{path}`.")


class DockerDaemonConnectionError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(f"Cannot connect to Docker Daemon, this may indicate that it is not running. {message}")


class NotSupportedError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(f"Not Supported: {message}")


# OS Exceptions
class PrivilegeError(Exception):
    pass


class InterfaceNotFoundError(Exception):
    pass


class HostArchitectureError(Exception):
    def __init__(self, architecture: str) -> None:
        super().__init__(f"Not implemented for host architecture `{architecture}`.")


# Lab Exceptions
class LabAlreadyExistsError(Exception):
    pass


class LabNotFoundError(Exception):
    pass


class EmptyLabError(Exception):
    def __init__(self) -> None:
        super().__init__("No devices in the current network scenario.")


class MachineDependencyError(Exception):
    pass


# Machine Exceptions
class MountDeniedError(Exception):
    pass


class MachineAlreadyExistsError(Exception):
    def __init__(self, machine_name: str) -> None:
        super().__init__(f"Device with name `{machine_name}` already exists.")


class NonSequentialMachineInterfaceError(Exception):
    def __init__(self, iface_num: int, machine_name: str) -> None:
        super().__init__(f"Interface `{iface_num}` missing on device `{machine_name}`.")


class MachineOptionError(Exception):
    pass


class MachineCollisionDomainError(Exception):
    pass


class MachineNotFoundError(Exception):
    pass


class MachineNotRunningError(Exception):
    def __init__(self, machine_name: str) -> None:
        super().__init__(f"Device `{machine_name}` is not running.")


class MachineNotReadyError(Exception):
    def __init__(self, machine_name: str) -> None:
        super().__init__(f"Device `{machine_name}` is not ready.")


class MachineBinaryError(Exception):
    __slots__ = ['binary', 'machine_name']

    def __init__(self, binary: str, machine_name: str):
        self.binary: str = binary
        self.machine_name: str = machine_name

    def __str__(self):
        return f"Binary `{self.binary}` not found in device `{self.machine_name}`."


# Interface Exceptions
class InterfaceMacAddressError(Exception):
    def __init__(self, mac_address: str, interface_num: int, machine_name: str) -> None:
        super().__init__(
            f"MAC address {mac_address} on interface `{interface_num}` of device `{machine_name}` is invalid."
        )


# Link Exceptions
class LinkNotFoundError(Exception):
    pass


class LinkAlreadyExistsError(Exception):
    pass


# Test Exceptions
class TestError(Exception):
    pass


class MachineSignatureNotFoundError(TestError):
    __slots__ = ['machine_name']

    def __init__(self, machine_name: str):
        self.machine_name: str = machine_name

    def __str__(self):
        return f"Signature for device `{self.machine_name}` not found!"


# Docker Exceptions
class InvalidImageArchitectureError(ValueError):
    __slots__ = ['image_name', 'arch']

    def __init__(self, image_name: str, arch: str):
        self.image_name: str = image_name
        self.arch: str = arch

    def __str__(self):
        return f"Docker Image `{self.image_name}` is not compatible with your host architecture `{self.arch}`"


class DockerImageNotFoundError(Exception):
    def __init__(self, image_name: str) -> None:
        super().__init__(f"Docker Image `{image_name}` is not available neither on Docker Hub nor in local repository!")


class DockerPluginError(Exception):
    pass


# Kubernetes Exceptions
class KubernetesConfigMapError(Exception):
    pass



from typing import Optional, Dict, Any

from ...foundation.setting.SettingsAddon import SettingsAddon

DEFAULTS = {
    "api_server_url": "Empty String",
    "api_token": "Empty String",
    "host_shared": True,
    "image_pull_policy": "IfNotPresent",
    "docker_config_json": "Empty String"
}

DEFAULT_DOCKER_CONFIG_JSON_PATH = "~/.docker/config.json"


class KubernetesSettingsAddon(SettingsAddon):
    __slots__ = ['api_server_url', 'api_token', 'host_shared', 'image_pull_policy', 'docker_config_json']

    def __init__(self) -> None:
        self.api_server_url: Optional[str] = None
        self.api_token: Optional[str] = None
        self.host_shared: bool = True
        self.image_pull_policy: Optional[str] = "IfNotPresent"
        self.docker_config_json: Optional[str] = None

    def _to_dict(self) -> Dict[str, Any]:
        return {
            'api_server_url': self.api_server_url,
            'api_token': self.api_token,
            'host_shared': self.host_shared,
            'image_pull_policy': self.image_pull_policy,
            'docker_config_json': self.docker_config_json
        }



import base64
from typing import Optional

from kubernetes import client
from kubernetes.client.api import core_v1_api
from kubernetes.client.rest import ApiException

from ...exceptions import KubernetesConfigMapError
from ...model.Machine import Machine
from ...utils import human_readable_bytes

MAX_FILE_SIZE = 3145728


class KubernetesConfigMap(object):
    """Class responsible for interacting with Kubernetes ConfigMap."""
    __slots__ = ['client']

    def __init__(self) -> None:
        self.client: core_v1_api.CoreV1Api = core_v1_api.CoreV1Api()

    def deploy_for_machine(self, machine: Machine) -> Optional[client.V1ConfigMap]:
        """Deploy and return a Kubernetes ConfigMap for the device.

        Args:
            machine (Kathara.model.Machine.Machine): A Kathara device.

        Returns:
            Optional[client.V1ConfigMap]: A Kubernetes ConfigMap.
        """
        config_map = self._build_for_machine(machine)

        if config_map is None:
            return None

        return self.client.create_namespaced_config_map(body=config_map, namespace=machine.lab.hash)

    def delete_for_machine(self, machine_name: str, machine_namespace: str) -> None:
        """Delete the Kubernetes ConfigMap associated with the device, if it exists.

        Args:
            machine_name (str): The name of a Kathara device.
            machine_namespace (str): the name of the namespace the device belongs to.

        Returns:
            None
        """
        try:
            self.client.delete_namespaced_config_map(name=self.build_name_for_machine(machine_name, machine_namespace),
                                                     namespace=machine_namespace
                                                     )
        except ApiException:
            return

    @staticmethod
    def build_name_for_machine(machine_name: str, machine_namespace: str) -> str:
        """Return the name for the Kubernetes deployment.

        Args:
            machine_name (str): The name of a Kathara device.
            machine_namespace (str): the name of the namespace the device belongs to.

        Returns:
            str: The name for the ConfigMap in the format '|machine_name|-|machine_namespace|-files'.
        """
        return "%s-%s-files" % (machine_name, machine_namespace)

    def _build_for_machine(self, machine: Machine) -> Optional[client.V1ConfigMap]:
        """Build and return a Kubernetes ConfigMap for the device.

        Args:
            machine (Kathara.model.Machine.Machine): A Kathara device.

        Returns:
            Optional[client.V1ConfigMap]: The Kubernetes ConfigMap for the device.

        Raises:
            KubernetesConfigMapError: If the device folder size exceeds the maximum supported size.
        """
        tar_data = machine.pack_data()

        # Create a ConfigMap on the cluster containing the base64 of the .tar.gz file
        # This will be decoded and extracted in the postStart hook of the pod
        if tar_data:
            # Before creating the .tar.gz file, check if it is bigger than the maximum allowed size.
            tar_data_size = len(tar_data)
            if tar_data_size > MAX_FILE_SIZE:
                raise KubernetesConfigMapError(
                    'Unable to upload device folder. Maximum supported size: %s. Current: %s.' % (
                        human_readable_bytes(MAX_FILE_SIZE),
                        human_readable_bytes(tar_data_size)
                    )
                )

            data = {"hostlab.b64": base64.b64encode(tar_data).decode('utf-8')}
            metadata = client.V1ObjectMeta(name=self.build_name_for_machine(machine.meta['real_name'],
                                                                            machine.lab.hash),
                                           deletion_grace_period_seconds=0
                                           )

            config_map = client.V1ConfigMap(api_version="v1",
                                            kind="ConfigMap",
                                            data=data,
                                            metadata=metadata
                                            )

            return config_map

        return None



import logging
from typing import Dict, Any

import requests

from ..exceptions import HTTPConnectionError

GITHUB_RELEASES_URL: str = "https://api.github.com/repos/%s/releases/latest"
REPOSITORY_NAME: str = "KatharaFramework/Kathara"
REQUEST_TIMEOUT: int = 1


class GitHubApi(object):
    @staticmethod
    def get_release_information() -> Dict[str, Any]:
        try:
            result = requests.get(GITHUB_RELEASES_URL % REPOSITORY_NAME, timeout=(REQUEST_TIMEOUT, REQUEST_TIMEOUT))
        except requests.exceptions.Timeout as e:
            raise HTTPConnectionError(str(e))
        except requests.exceptions.ConnectionError as e:
            raise HTTPConnectionError(str(e))

        if result.status_code != 200:
            logging.debug("GitHub replied with status code %s while looking for Kathara repo.", result.status_code)
            raise HTTPConnectionError("GitHub replied with status code %s." % result.status_code)

        return result.json()



# From rcludwick/depgen.py
# These functions take a dictionary of dependencies in the following way:
# depdict = { 'a' : [ 'b', 'c', 'd'],            'b' : [ 'c', 'd'],            'e' : [ 'f', 'g']            }
# has_loop() will check for dep loops in the dep dict with true or false.
# flatten() will create an ordered list of items according to the dependency structure.
# Note: To generate a list of dependencies in increasing order of dependencies, say for a build, run: flatten(MyDepDict)
import types


def _order(idepdict, val=None, level=0):
    """
    Generates a relative order in a dep dictionary
    :param idepdict: 
    :param val: 
    :param level: 
    :return: 
    """
    results = {}
    if val is None:
        for (k, v) in idepdict.items():
            for dep in v:
                results.setdefault(k, 0)
                d = _order(idepdict, val=dep, level=level+1)
                for dk, dv in d.items():
                    if dv > results.get(dk, 0):
                        results[dk] = dv

        return results
    else:
        results[val] = level
        deps = idepdict.get(val, None)
        if deps is None or deps == []:
            return {val: level}
        else:
            for dep in deps:
                d = _order(idepdict, val=dep, level=level+1)
                for dk, dv in d.items():
                    if dv > results.get(dk, 0):
                        results[dk] = dv

            return results


def _invert(d):
    """
    Inverts a dictionary
    :param d: 
    :return: 
    """
    i = {}
    for (k, v) in d.items():
        if isinstance(v, list):
            for dep in v:
                depl = i.get(dep, [])
                depl.append(k)
                i[dep] = depl
        else:
            depl = i.get(v, [])
            depl.append(k)
            i[v] = depl

    return i


def flatten(depdict):
    """
    flatten() generates a list of deps in order
    :param depdict: 
    :return: 
    """
    # Generate an inverted deplist
    ideps = _invert(depdict)

    # Generate relative order
    order = _order(ideps)

    # Invert the order
    iorder = _invert(order)

    # Sort the keys and append to a list
    output = [] 
    for key in sorted(iorder.keys()):
        output.extend(iorder[key])

    return output


def has_loop(depdict, seen=None, val=None):
    """
    Check to see if a given depdict has a dependency loop
    :param depdict: 
    :param seen: 
    :param val: 
    :return: 
    """
    if seen is None:
        for k, v in depdict.items(): 
            seen = []
            for val in v: 
                if has_loop(depdict, seen=list(seen), val=val):
                    return True
    else:
        if val in seen:
            return True
        else:
            seen.append(val)
            k = val
            v = depdict.get(k, [])
            for val in v:
                if has_loop(depdict, seen=list(seen), val=val):
                    return True

    return False 


from collections.abc import Iterator
from typing import Generator

from docker import DockerClient

from ....foundation.manager.exec_stream.IExecStream import IExecStream


class DockerExecStream(IExecStream):
    """Docker-specific class for handling the commands stream exec.

    Attributes:
        _stream (Generator): The generator yielding the output of the stream exec.
        _stream_api_object (Any): The specific API object to interact with the underlying stream.
        _client (DockerClient): The Docker client object to interact with the stream.
    """
    __slots__ = ['_client']

    def __init__(self, stream: Generator, stream_api_object: str, client: DockerClient) -> None:
        super().__init__(stream, stream_api_object)

        self._client: DockerClient = client

    def stream_next(self) -> Iterator:
        """Return the next element from the stream.

        Returns:
            Iterator: The output iterator from the stream.
        """
        return next(self._stream)

    def exit_code(self) -> int:
        """Return the exit code of the execution.

        Returns:
            int: The exit code of the execution.
        """
        return int(self._client.api.exec_inspect(self._stream_api_object)['ExitCode'])



from typing import Dict, Any, List

from docker.errors import NotFound
from docker.models.containers import Container
from docker.models.networks import Network

from ....foundation.manager.stats.ILinkStats import ILinkStats


class DockerLinkStats(ILinkStats):
    """The class responsible to handle Docker Networks statistics.

    Attributes:
        link_api_object (Network): The Docker Network associated with this statistics.
        lab_hash (str): The hash identifier of the network scenario of the Docker Network.
        name (str): The name of the collision domain.
        network_name (str): The Docker Network Name.
        user (str): The user that deployed the associated Docker Network.
        enable_ipv6 (bool): True if ipv6 is enabled, else None.
        external (List[str]): A list with the name of the attached external networks.
        containers (List[Container]): A list of the Docker Container associated with the Docker Network.
    """
    __slots__ = ['link_api_object', 'lab_hash', 'name', 'network_name', 'user', 'enable_ipv6', 'external', 'containers']

    def __init__(self, link_api_object: Network):
        self.link_api_object: Network = link_api_object
        self.lab_hash: str = link_api_object.attrs.get('Labels')['lab_hash']
        self.name: str = link_api_object.attrs.get('Labels')['name']
        self.network_name: str = link_api_object.name
        self.user: str = link_api_object.attrs.get('Labels')['user']
        self.enable_ipv6: bool = link_api_object.attrs.get('EnableIPv6')
        external = link_api_object.attrs.get('Labels')['external']
        self.external: List[str] = external.split(";") if external else []
        self.containers: List[Container] = []
        self.update()

    def update(self) -> None:
        """Update dynamic statistics with the current ones.

        Returns:
            None
        """
        try:
            self.link_api_object.reload()
        except NotFound:
            # Happens while deleting
            pass

        self.containers = [container for container in self.link_api_object.containers]

    def to_dict(self) -> Dict[str, Any]:
        """Transform statistics into a dict representation.

        Returns:
            Dict[str, Any]: Dict containing statistics.
        """
        return {
            "network_scenario_id": self.lab_hash,
            "name": self.name,
            "network_name": self.network_name,
            "user": self.user,
            "enable_ipv6": self.enable_ipv6,
            "external": self.external,
            "containers": self.containers
        }

    def __repr__(self) -> str:
        return str(self.to_dict())

    def __str__(self) -> str:
        """Return a formatted string with the link statistics.

        Returns:
           str: A formatted string with the link statistics
        """
        formatted_stats = f"Network Scenario ID: {self.lab_hash}"
        formatted_stats += f"\nLink Name: {self.name}"
        formatted_stats += f"\nNetwork Name: {self.network_name}"
        formatted_stats += f"\nEnable IPv6: {self.enable_ipv6}"
        if self.external:
            formatted_stats += f"\nExternal Interfaces:"
            for ext in self.external:
                formatted_stats += f"\n\t- {ext}\n"
        if self.containers:
            formatted_stats += f"\nAttached Devices:"
            for container in self.containers:
                formatted_stats += f"\n\t- {container.labels['name']}"

        return formatted_stats
